/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "c_context.h"

/* We don't use getcontext() anymore because some machines (I saw this on 
   GNU/Linux Sparc) don't have it. We can do the same with a setjmp(). */
//#define USE_GETCONTEXT

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef USE_GETCONTEXT
#include <ucontext.h>
#else
#include <setjmp.h>
#endif // #ifdef USE_GETCONTEXT
#include <assert.h>

/* We hope no processor really has such a stack bottom... */
#define uninitialized_stack_bottom_approximation ((void**)-1)

/* The address of a variable allocated at the 'stack bottom'; this is not
   actually the bottom, but it's in a frame below the one used for the
   user-supplied functions: */
static void **stack_bottom_approximation =
  uninitialized_stack_bottom_approximation;

/* Set our stack bottom pointer approximation to a user-supplied value.
   This is useful when the user can't initialize in our funny
   'continuation' style by supplying a pointer to the function used to
   execute the rest of the program. */
void eam_save_stack_bottom(void *stack_bottom){
  stack_bottom_approximation = (void**)stack_bottom;
}

/* The returned value has no meaning. We only use it to trick GCC into not
   optimizing recursive calls. */
static int push_frames_and_call(const int n,
                                void *unused,
                                void (*function)(const void *stack_bottom,
                                                 const void *stack_top)){
  /* This is useful for debugging; the volatile modifier might also prevent
     GCC from using a single frame for all recursive activations: */
  volatile void* an_object_in_this_frame = (void*)0x1234abcd;
  
  if(n == -1){
    /* Unreachable code. This prevents GCC from optimizing away the variable
       'an_object_in_this_frame' and the parameter 'unused': */
    fprintf(stderr, 
            "VERY SERIOUS WARNING: this code should never be reached.\n");
    return (int)(an_object_in_this_frame - unused);
  }
  else if(n == 0){
    /* We take the stack as ranging from our 'stack bottom' to the address of
       this local variable: */
    volatile void* object_at_the_stack_top = (void*)0xffffffff;
    
    /*
    fprintf(stderr, "[[[ BOTTOM is %p; TOP is %p]]]\n",
            (void*) stack_bottom_approximation,
            (void*) &object_at_the_stack_top);
    */
    /* Ok, we pushed enough frames on the stack. Now do the work: the returned
       value has the only purpose of fooling GCC, preventing it from making 
       tail-call optimizations: */
    function((void*) stack_bottom_approximation,
             (void*) &object_at_the_stack_top);
    return (int)((void*)&object_at_the_stack_top
                 - (void*)stack_bottom_approximation);
  }
  else
    /* We want to push a new frame on the stack, and to obtain this we
       make a recursive call. Note that we *don't* want this to be
       optimized into a tail-call! This trivial arithmetic on the
       returned value should be enough to prevent GCC from playing
       too clever tricks (tested up to GCC 4.2). */
    return 1 + push_frames_and_call(n - 1,
                                    &an_object_in_this_frame,
                                    function) / 3;
}

static const int nested_frames_no = 20;//20; // To do: make this smaller?
void eam_call_with_stack_bottom_and_stack_top(void (*function)(const void *stack_bottom,
                                                           const void *stack_top)){
#ifdef ENABLE_ASSERTIONS
  if(stack_bottom_approximation == uninitialized_stack_bottom_approximation){
    fprintf(stderr, "FATAL: the stack bottom pointer was not initialized\n");
    assert(0);
  } // if
#endif // #ifdef ENABLE_ASSERTIONS
  /* We want to be sure that what we mean as 'top' is acutally above the
     current frame: so we make some recursive calls to push a few more
     frames; a local variable into the topmost frame will be our 'top': */
  push_frames_and_call(nested_frames_no,
                       //*stack_bottom_approximation, // ??? To do: The "*" is suspicious
                       stack_bottom_approximation, // ??? To do: The "*" is suspicious
                       function);
}

/* Initialize the stack bottom and then call the user-supplied function on
   the given arguments. The value returned from the user-supplied function
   is returned. Note that the stack bottom is guaranteed to be correct only
   in the context of the function. */
int eam_save_stack_bottom_and_call_main(
       int (*main_function_from_the_user)(int argc, char **argv),
       int argc,
       char **argv){
  /* This is essentially at the stack bottom: the main function supplied by
     the user will be evaluated in a frame spawned by the current one. */
  void* something_at_the_stack_bottom = (void*)0xaaaaaaaa;
  
  /* Set the stack bottom: it will remain correct until 
     main_function_from_the_user() returns. When debugging we enforce
     this restriction by explicitly invalidating the bottom pointer: */
  stack_bottom_approximation = &something_at_the_stack_bottom;
  
  /* Instead of just returning the result of applying the user-supplied
     function to argc and argv, which could be optimized into a sibling-call
     (and we don't want it: user variables should be *above* our stack
     bottom; the current GCCs don't do that for functions called via a
     pointer, but this may change), we trick GCC into believing the
     assignment to useless_variable is a reason good enough to prevent
     optimization: :-) */
  int return_value = main_function_from_the_user(argc, argv);
  volatile int useless_variable;
  useless_variable = return_value;
#ifdef ENABLE_ASSERTIONS
  stack_bottom_approximation = uninitialized_stack_bottom_approximation;
#endif // #ifdef ENABLE_ASSERTIONS
  return useless_variable;
}

/* The idea is the same as for eam_save_stack_bottom_and_call_main(). Yes, there
   is some code duplication; it's difficult to factorize code like this in
   C, when there are pointers to functions with different argument types. */
static volatile int useless_variable = 0;
void eam_save_stack_bottom_and_continue(void (*function_from_the_user)(void)){
  void* something_at_the_stack_bottom = (void*)0xaaaaaaaa;
  stack_bottom_approximation = &something_at_the_stack_bottom;
  function_from_the_user();
  useless_variable++;
#ifdef ENABLE_ASSERTIONS
  stack_bottom_approximation = uninitialized_stack_bottom_approximation;
#endif // #ifdef ENABLE_ASSERTIONS
}

void eam_call_with_first_register_and_last_register(void (*function)(const void *first_register,
                                                                 const void *last_register)){
  /* Save the current context, which includes an image of assembly
     registers: */
  struct context_plus_two_words{
    void *initial_padding;
#ifdef USE_GETCONTEXT
    ucontext_t actual_context;
#else
    jmp_buf actual_context;
#endif // #ifdef USE_GETCONTEXT
    void *final_padding;
  } context;
#ifdef USE_GETCONTEXT
  int result_of_getcontext = getcontext(&(context.actual_context));
  assert(result_of_getcontext == 0);
#else
  int result_of_setjmp = setjmp(context.actual_context);
  assert(result_of_setjmp == 0);
#endif // #ifdef USE_GETCONTEXT
  
  /* Call the given function, passing the address of the first word and of the
     last word as parameters: */
  function((void*)&context,
           (void*)(((char*)&context) +
                   (sizeof(struct context_plus_two_words) - sizeof(void*))));
}


#if 0
/* A simple test: */
void do_something_with_the_stack(const void *stack_bottom,
                                 const void *stack_top){
  const int distance = (int)(stack_top - stack_bottom);
  printf("bottom: %p [%u]\n", stack_bottom, (unsigned int)stack_bottom);
  printf("top:    %p [%u]\n", stack_top, (unsigned int)stack_top);
  printf("distance (in bytes): %i\n", distance);

  void *p;
  void *from, *to;
  if(stack_bottom < stack_top){
    from = (void*)stack_bottom;
    to = (void*)stack_top;
  } else{
    from = (void*)stack_top;
    to = (void*)stack_bottom;
  }

  for(p = from; p <= to; p += sizeof(void*))
    printf("%x ", *(int*)p);
  printf("\n");

}

void do_something_with_the_registers(const void *first_register,
                                     const void *last_register){
  void *p;
  for(p = (void*)first_register; p <= last_register; p += sizeof(void*))
    printf("%x ", *(int*)p);
  printf("\n");

}

int what_the_user_thinks_to_be_main(int argc, char **argv){
  eam_call_with_stack_bottom_and_stack_top(do_something_with_the_stack);
  printf("---------------------------------------------------\n");
  eam_call_with_stack_bottom_and_stack_top(do_something_with_the_registers);
  return 0;
}

void f(void){
  eam_call_with_stack_bottom_and_stack_top(do_something_with_the_stack);
  printf("---------------------------------------------------\n");
  eam_call_with_first_register_and_last_register(do_something_with_the_registers);
}

int main(int argc, char **argv){
  /*
  return eam_save_stack_bottom_and_call_main(what_the_user_thinks_to_be_main,
                                        argc, argv);
  */
  eam_save_stack_bottom_and_continue(f);
  return 0;
}
#endif
