/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005 Luca Saiu
Updated in 2006 by Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "list.h"
#include "malloc.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>



/* A list is just a pair. The user must take care of types, and of
   deallocation: */
struct eam_list{
  void *car;
  struct eam_list *cdr;
}; // struct

struct eam_list *eam_nil = NULL;

int eam_null(const struct eam_list *list){
  return list == NULL;
}

int eam_neam_null(const struct eam_list *list){
  return list != NULL;
}

struct eam_list* eam_cons(void *object, struct eam_list* list){
  struct eam_list *new_cons = (struct eam_list*)eam_xmalloc(sizeof(struct eam_list));
  new_cons->car = object;
  new_cons->cdr = list;
  
  return new_cons;
}

void* eam_car(const struct eam_list* list){
#ifdef ENABLE_ASSERTIONS
  assert(list != NULL);
#endif
  return list->car;
}

struct eam_list* eam_cdr(const struct eam_list* list){
#ifdef ENABLE_ASSERTIONS
  assert(list != NULL);
#endif
  return list->cdr;
}

void set_eam_car(struct eam_list* list, void *object){
#ifdef ENABLE_ASSERTIONS
  assert(list != NULL);
#endif
  list->car = object;
}

void set_eam_cdr(struct eam_list* list, struct eam_list *object){
#ifdef ENABLE_ASSERTIONS
  assert(list != NULL);
#endif
  list->cdr = object;
}

int eam_length(const struct eam_list* list){
  /* Scan the list, keeping track of the number of moves: */
  const struct eam_list *p = list;
  int current_length = 0;
  while(p != NULL){
    p = p->cdr;
    current_length++;
  } // while
  
  /* Just return the number we kept: */
  return current_length;
}

static void destroy_list(const struct eam_list *list, const int destroy_elements) 
  __attribute__(( always_inline ));
static void destroy_list(const struct eam_list *list, const int destroy_elements){
  const struct eam_list *p = list;
  while(p != NULL){
    /* Destroy the cons and, if requested, the car; then do the same on the cdr,
       until the cdr is eam_nil: */
    const void *car = p->car;
    const struct eam_list *cdr = p->cdr;
    free((void*)p);
    if(destroy_elements)
      free((void*)car);
    
    /* Go on with the next element: */
    p = cdr;
  } // while
}

static struct eam_list* reverse_with_accumulator(const struct eam_list *list,
					     struct eam_list *acc){
  if(eam_null(list))
    return acc;
  else
    return reverse_with_accumulator(eam_cdr(list), eam_cons(eam_car(list), acc));
}

struct eam_list* eam_reverse(const struct eam_list *list){
  return reverse_with_accumulator(list, (struct eam_list*)eam_nil);
}

void eam_destroy_list_structure(const struct eam_list *list){
  destroy_list(list, 0);
}

void eam_destroy_whole_list(const struct eam_list *list){
  destroy_list(list, 1);
}

#if 0
/* A simple test: */
struct eam_list* interval(const int a, const int b){
  if(a > b)
    return eam_nil;
  else{
    int *x = eam_xmalloc(sizeof(int));
    *x = a;
    return eam_cons(x, interval(a + 1, b));
  }
}

int main(){
  struct eam_list *xs = interval(1, 1000);
  struct eam_list *ys = eam_reverse(xs);

  eam_destroy_list_structure(ys);
  eam_destroy_whole_list(xs);
  return 0;
}
#endif
