/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005 Luca Saiu
Minor updates in 2006 by Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include "../common/types.h"
#include "../common/command_line.h"
#include "../common/malloc.h"
#include "mangle.h"

char* eam_mangle(const char *identifier){
  /* In the worst case the size of the mangled identifier is four times the
     size of the unmangled identifier: the most expensive escaping is the
     case _DDD, where a character is mapped into an underscore and three
     digits. We have to take into account also the initial prefix and the
     trailing '\0': */
  const int identifier_length = strlen(identifier);
  char *buffer = (char*)eam_xmalloc(identifier_length * 4 + 2);
  /* Set the constant buffer prefix: */
  int i, buffer_i = 1;
  buffer[0] = 'e';
  
  /* Translate each character: */
  for(i = 0; i < identifier_length; i++){
    if(isdigit(identifier[i]))
      /* Digit: no escape: */
      buffer[buffer_i++] = identifier[i];
    else if(isascii(identifier[i]) && islower(identifier[i]))
      /* Lower-case ASCII letter: no escape: */
      buffer[buffer_i++] = identifier[i];
    else if(isascii(identifier[i]) && isupper(identifier[i])){
      /* We need an escape for an upper-case ASCII letter: */
      buffer[buffer_i] = '_';
      buffer[buffer_i + 1] = tolower(identifier[i]);
      buffer_i += 2;
    }
    else if(identifier[i] == '_'){
      /* We need an escape for '_': */
      buffer[buffer_i] = '_';
      buffer[buffer_i + 1] = '_';
      buffer_i += 2;
    }
    else{
      /* We need an escape for the generic case: */
      buffer[buffer_i] = '_';
      buffer[buffer_i + 1] = '0' + (char)(((int)(identifier[i])) / 100);
      buffer[buffer_i + 2] = '0' + (char)(((int)(identifier[i])) % 100 / 10);
      buffer[buffer_i + 3] = '0' + (char)(((int)(identifier[i])) % 10);
      buffer_i += 4;
    }
  } // for  
  
  /* Terminate the mangled copy: */
  buffer[buffer_i] = '\0';
    
  /* Return the buffer we filled, possibly shrunk to save space. */
  return (char*)realloc(buffer, buffer_i + 1);
}

char* eam_unmangle(const char *identifier){
  /* The size of an unmangled identifier is less or equal to the size of its
     mangled form. We also have to take into account the trailing '\0': */
  const int identifier_length = strlen(identifier);
  char *buffer = (char*)eam_xmalloc(identifier_length + 1);
  int i, buffer_i = 0;
  /* Check the constant buffer prefix: */
  if(identifier[0] != '_')
    eam_fatal("invalid mangled string prefix");
  
  /* Translate each character: */
  for(i = 1; i < identifier_length; i++){
    if(isdigit(identifier[i]))
      /* This is a digit, no escape: */
      buffer[buffer_i++] = identifier[i];
    else if(isascii(identifier[i]) && islower(identifier[i]))
      /* This is an ASCII lower-case letter, no escape: */
      buffer[buffer_i++] = identifier[i];
    else if(identifier[i] != '_'){
      /* A mangled identifier can only contain alphanumerics and escapes: */
      eam_fatal("invalid non-escaped charater");
    }
    /* If we're here then we found an escape sequence: translate it: */
    else if(i == identifier_length - 1)
      eam_fatal("truncated trailing single character escape sequence");
    else if(identifier[i + 1] == '_'){
      /* Escape for '_': */
      buffer[buffer_i++] = '_';
      i++; // skip the second '_'
    }
    else if(isascii(identifier[i + 1]) && islower(identifier[i + 1])){
      /* Escape for an upper-case ASCII letter: */
      buffer[buffer_i++] = toupper(identifier[i + 1]);
      i++; // skip the second character
    }
    /* If we're here then we found a generic character escape: */
    else if(i > identifier_length - 4)
      eam_fatal("truncated trailing generic escape sequence");
    else if((!isdigit(identifier[i + 1])) || 
	    (!isdigit(identifier[i + 2])) || 
	    (!isdigit(identifier[i + 3]))){
      eam_fatal("invalid generic escape sequence");
    }
    else{
      /* We found a valid generic escape sequence: */
      const char c = (identifier[i + 1] - '0') * 100 +
                     (identifier[i + 2] - '0') * 10 +
                     (identifier[i + 3] - '0');
      buffer[buffer_i++] = c;
      i += 3; // skip the three digit characters
    }
  } // for  
  
  /* Terminate the unmangled copy: */
  buffer[buffer_i] = '\0';
  
  /* Return the buffer we filled, possibly shrunk to save space. */
  return (char*)realloc(buffer, buffer_i + 1);
}

static uint64_t integer_exp(const uint64_t b, const uint64_t e){
  uint64_t i, r = 1u;
  for(i = 0; i < e; i++)
    r *= b;
  return r;
}

/* An hash function, returing an unsigned number, truncated to its
   rightmost n base-37 digits: */
static uint64_t hash(const char *string, const int n){
  const int string_size = strlen(string);
  /* We use uint64_t because we want predictable and machine-independent
     wrapping semantics: */
  uint64_t r = 0;
  int i;
  for(i = 0; i < string_size; i++)
    r += integer_exp(((uint64_t)(i + 127u)) % 23u + 11u, 
		     ((uint64_t)(string[i]) + 12u) % 23u + 12u);
  
  /* Return r modulo 37^n, so that we use at most n characters in the
     printed base-37 representation: */
  return r % integer_exp(37u, n);
}

/* Return the given integer (which should be in [0, 36]) encoded as an
   "acceptable" character, i.e. one of the one included in mangled
   identifiers: */
static char encode_as_character(int c){
  assert(c >= 0);
  assert(c <= 36);
  if(c == 0)
    return '_';
  else if(c < 11)
    return (char)(c - 1 + '0');
  else
    return (char)(c - 11 + 'a');
}

char* eam_mangle_with_limit(const char *identifier, const int n){
  /* Get the malloced() mangled string, not limited in length: */
  char* mangled_identifier = eam_mangle(identifier);
  
  /* Just return the mangled string if it isn't too long: */
  if(strlen(mangled_identifier) <= n)
    return mangled_identifier;
  
  /* The mangled string is too long. We (destructively) replace it with
     an hash of the unmangled identifier: */
  mangled_identifier[0] = 'h'; // prefix for hashed mangled names
  uint64_t hashed_value = hash(identifier, n - 1);
  int mangled_identifier_i = 1;
  while(hashed_value != 0){
    mangled_identifier[mangled_identifier_i++] =
      encode_as_character((int)(hashed_value % (uint64_t)37u));
    /*
    fprintf(stderr,
	    "%-3i -> %c\n",
	    (int)(hashed_value % (uint64_t)37u),
	    encode_as_character((int)(hashed_value % (uint64_t)37u)));
    */
    hashed_value /= (uint64_t)37u;
  } // while
  
  /* Terminate the string and return it, possibly shrunk to save space. */
  mangled_identifier[mangled_identifier_i] = '\0';
  return (char*)realloc(mangled_identifier, mangled_identifier_i + 1);
}

#if 0
#define TESTS_NO 10000000
#define N 200
void test(const char* unmangled){
  char *mangled = eam_mangle(unmangled);
  char *unmangled2 = uneam_mangle(mangled);
  assert(! strcmp(unmangled, unmangled2));
  free(mangled);
  free(unmangled2);
}

int run_tests(){
  int i, j;
  char s[N];
  printf("Starting tests...\n");
  for(i = 0; i < TESTS_NO; i++){
    int size = rand() % N;
    for(j = 0; j < size; j++)
      s[j] = (char)(rand() % 256);
    s[size] = '\0';
    test(s);
  } // outer for
  printf("...all tests were passed.\n");
  return 0;
}

int main(){
  char *s = "TrallalLero trallalla` e aNcOra qUAlcosa iN piu`, gIusto per avere una stringa lunga";
  printf(">%s<\n== mangle ==>\n", s);
  printf(">%s<\n\n", eam_mangle(s));
  printf(">%s<\n== mangle-with-limit ==>\n", s);
  printf(">%s<\n", eam_mangle_with_limit(s, 15));
  return 0;
}
#endif
