/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#ifndef EAM_MANGLE_H_
#define EAM_MANGLE_H_

#include "../common/types.h"

/* Mangling allows to map high-level names into identifiers which are directly
   usable by the compiler, the linker and even the filesystem.
   Mangling schema:
   - a decimal digit like '3' is mapped into itself
   - an ASCII lower-case letter like 'a' is mapped into itself
   - an ASCII upper-case letter like 'A' is mapped into "_a"
   - the underscore '_' is mapped into "__"
   - any other character, like '?', is mapped into "_AAA", where AAA is the
     *decimal* ASCII encoding for '?'. We always use three digits.
   
   A prefix "e" is prepended to mangled names.
   This mapping is injective and thus reversable; both directions are
   implemented. */

/* Return a malloc()ed string containing the given identifier in mangled
   form. This is useful so that we can : */
char* eam_mangle(const char *identifier);

/* Return a malloc()ed string containing the given mangled identifier in
   unmangled form; fail with a fatal error if identifier is not a valid
   mangled identifier: */
char* eam_unmangle(const char *identifier);

/* Return a malloc()ed string containing: 
   - a mangled version of identifier, if the mangled string length is less than
   or equal to n;
   - otherwise the result of applying an hash function to identifier, with an
   'h' character prepended, so that the total length of the result is no more
   than n characters, excluding the trailing '\0'. Using the hash breaks the 
   injectivity of the mangling function, but allows to use mangling when the
   mangled identifier size is limited. This should not cause problems in
   practice, and we feel our behaviour is less dangerous than silently
   truncating strings. */
char* eam_mangle_with_limit(const char *identifier, const int n);

/* The number of characters used to eam_mangle_with_limit() compilation unit names;
   the length of the representation of the name of defined *objects* is not
   incuded here. */
#define EAML_COMPILATION_UNIT_MANGLED_NAME_LENGTH 120 //12

/* The number of characters used to represent an exported symbol. An exported
   symbol is made by an initial lower-case letter identifying the object sort,
   followed by a uniqe identification number, greater than or equal to zero. */
#define EAML_SYMBOL_NAME_LENGTH 6 // To do: kill this

#endif // #ifndef EAM_MANGLE_H_
