#include "../../src/common/types.h"
#include "../../src/common/string_map.h"
#include "../../src/common/list.h"
#include "../../src/libraries/gc/gc.h"
#include "../../src/libraries/gc/generic_pools.h"
#include <string.h>

enum scheme_type{
  empty_list_scheme_type = 1, // only one object
  number_scheme_type = 2,
  boolean_scheme_type = 3, // only two objects
  character_scheme_type = 4,
  symbol_scheme_type = 5,
  string_scheme_type = 6,
  cons_scheme_type = 7,
  closure_scheme_type = 8,
  primitive_scheme_type = 9,
  variable_scheme_type = 10, // the user should never see this
  array_scheme_type = 0 // we use generic pools for this
}; // enum

/* A mapping from symbols to their (conventionally) unique string
   representation, usable for equality-by-identity comparisons. Unique
   string representations are chained together in the list. Note that 
   neither the map nor the list is subject to GC.
   GC'd symbol objects are tagged pointers to elements of
   scheme_symbols_list. */
struct eam_string_map *scheme_symbols_map;
struct eam_list *scheme_symbols_list;

/* Homogeneous pools for Scheme objects, whose type must be efficiently
   recognizable: */
struct eamgc_homogeneous_pool *empty_list_pool;
struct eamgc_homogeneous_pool *numbers_pool;
struct eamgc_homogeneous_pool *booleans_pool;
struct eamgc_homogeneous_pool *characters_pool;
struct eamgc_homogeneous_pool *symbols_pool;
struct eamgc_homogeneous_pool *strings_pool;
struct eamgc_homogeneous_pool *conses_pool;
struct eamgc_homogeneous_pool *closures_pool;
struct eamgc_homogeneous_pool *primitives_pool;
/* We use generic pools for arrays. */

/* Variables are just indirect references: */
struct eamgc_homogeneous_pool *variables_pool;

/* Unique objects: */
word_t *scheme_empty_list, *scheme_true, *scheme_false;

void actual_main(void);
void create_pools_for_scheme_objects(void);

void trace_second_element(const word_t object){
  eamgc_trace_nth_word(object, 1);
}

int main(){
  eamgc_initialize_garbage_collector_and_continue(actual_main);
  return 0;
}

void create_pools_for_scheme_objects(void){
  empty_list_pool =
    eamgc_create_homogeneous_pool(1,
                                  empty_list_scheme_type,
                                  NULL,
                                  eamgc_homogeneous_trace_nothing_function); 
  numbers_pool =
    eamgc_create_homogeneous_pool(1,
                                  number_scheme_type,
                                  NULL,
                                  eamgc_homogeneous_trace_nothing_function); 
  booleans_pool =
    eamgc_create_homogeneous_pool(1,
                                  boolean_scheme_type,
                                  NULL,
                                  eamgc_homogeneous_trace_nothing_function);
  characters_pool =
    eamgc_create_homogeneous_pool(1,
                                  character_scheme_type,
                                  NULL,
                                  eamgc_homogeneous_trace_nothing_function);
  symbols_pool =
    eamgc_create_homogeneous_pool(1,
                                  symbol_scheme_type,
                                  NULL,
                                  eamgc_homogeneous_trace_nothing_function);
  strings_pool =
    eamgc_create_homogeneous_pool(1,
                                  string_scheme_type,
                                  NULL,
                                  eamgc_homogeneous_trace_nothing_function);
  conses_pool =
    eamgc_create_homogeneous_pool(2,
                                  cons_scheme_type,
                                  NULL,
                                  eamgc_homogeneous_trace_everything_function);
  closures_pool =
    eamgc_create_homogeneous_pool(2,
                                  closure_scheme_type,
                                  NULL,
                                  trace_second_element);
  primitives_pool =
    eamgc_create_homogeneous_pool(1,
                                  primitive_scheme_type,
                                  NULL,
                                  eamgc_homogeneous_trace_nothing_function);
  variables_pool =
    eamgc_create_homogeneous_pool(1,
                                  variable_scheme_type,
                                  NULL,
                                  eamgc_homogeneous_trace_everything_function);
}

void create_symbols_structures(void){
  scheme_symbols_list = eam_nil;
  scheme_symbols_map = eam_create_string_map(sizeof(char*),
                                             eam_default_string_map_datum_finalizer);
}

void create_unique_scheme_objects(void){
  scheme_empty_list = eamgc_allocate_from_homogeneous_pool(empty_list_pool);
  *scheme_empty_list = NULL; // just to ease debugging: there's no need for this
  
  scheme_true = eamgc_allocate_from_homogeneous_pool(booleans_pool);
  *scheme_true = (word_t)-1;
  
  scheme_false = eamgc_allocate_from_homogeneous_pool(booleans_pool);
  *scheme_false = (word_t)0;
  
  /* Unique Scheme objects are GC roots: */
  eamgc_register_heap_root(scheme_empty_list);
  eamgc_register_heap_root(scheme_true);
  eamgc_register_heap_root(scheme_false);
}

static word_t make_single_word_object(struct eamgc_homogeneous_pool *pool,
                                      const word_t value){
  word_t *object = eamgc_allocate_from_homogeneous_pool(pool);
  *object = value;
  return (word_t)object;
}

word_t make_number(const integer_t x){
  return make_single_word_object(numbers_pool, (const word_t)x);
}

word_t make_character(const char x){
  return make_single_word_object(characters_pool,
                                 (const word_t)(const integer_t)x);
}

word_t make_variable(const word_t x){
  return make_single_word_object(variables_pool, x);
}

word_t make_symbol(const char* symbol_as_string){
  /* Make a new symbol and add it to structures if this is the first time
     we see it. Otherwise just use the symbol we already have: */
  char *copy_of_symbol_as_string = eam_access_string_map(scheme_symbols_map,
                                                         symbol_as_string);
  if(copy_of_symbol_as_string == NULL){
    copy_of_symbol_as_string = eam_xmalloc(strlen(symbol_as_string) + 1);
    strcpy(copy_of_symbol_as_string, symbol_as_string);
    fprintf(stderr, "Creating a symbol \"%s\"\n", copy_of_symbol_as_string);
    scheme_symbols_list = eam_cons(copy_of_symbol_as_string,
                                   scheme_symbols_list);
    eam_insert_into_string_map(scheme_symbols_map,
                               copy_of_symbol_as_string,
                               copy_of_symbol_as_string);
  } // if
  
  /* Now copy_of_symbol_as_string is our unique pointer representing the
     symbol: */
  return make_single_word_object(symbols_pool, 
                                 (const word_t)copy_of_symbol_as_string);
}

/*
  string_scheme_type = 6,
  closure_scheme_type = 8,
  primitive_scheme_type = 9,
  array_scheme_type = 0 // we use generic pools for this
*/

word_t make_cons(const word_t car, const word_t cdr){
  word_t *new_cons = eamgc_allocate_from_homogeneous_pool(conses_pool);
  new_cons[0] = car;
  new_cons[1] = cdr;
  return (word_t)new_cons;
}

word_t make_boolean(const integer_t c_boolean){
  if(c_boolean)
    return scheme_true;
  else
    return scheme_false;
}

/*
  empty_list_scheme_type = 1, // only one object
  number_scheme_type = 2,
  boolean_scheme_type = 3, // only two objects
  character_scheme_type = 4,
  symbol_scheme_type = 5,
  string_scheme_type = 6,
  cons_scheme_type = 7,
  closure_scheme_type = 8,
  primitive_scheme_type = 9,
  variable_scheme_type = 10, // the user should never see this
  array_scheme_type = 0 // we use generic pools for this
*/
integer_t is_empty_list(const word_t x){
  return eamgc_object_to_tag(x) == empty_list_scheme_type;
}
integer_t is_number(const word_t x){
  return eamgc_object_to_tag(x) == number_scheme_type;
}
integer_t is_boolean(const word_t x){
  return eamgc_object_to_tag(x) == boolean_scheme_type;
}
integer_t is_character(const word_t x){
  return eamgc_object_to_tag(x) == character_scheme_type;
}
integer_t is_symbol(const word_t x){
  return eamgc_object_to_tag(x) == symbol_scheme_type;
}
integer_t is_string(const word_t x){
  return eamgc_object_to_tag(x) == string_scheme_type;
}
integer_t is_cons(const word_t x){
  return eamgc_object_to_tag(x) == cons_scheme_type;
}
integer_t is_closure(const word_t x){
  return eamgc_object_to_tag(x) == closure_scheme_type;
}
integer_t is_primitive(const word_t x){
  return eamgc_object_to_tag(x) == primitive_scheme_type;
}
integer_t is_variable(const word_t x){
  return eamgc_object_to_tag(x) == variable_scheme_type;
}
integer_t is_array(const word_t x){
  return eamgc_object_to_tag(x) == array_scheme_type;
}

void print_scheme_object(word_t x, FILE *stream){
  switch(eamgc_object_to_tag(x)){
  case empty_list_scheme_type:{
    fprintf(stream, "()");
    break;
  }
  case number_scheme_type:{
    fprintf(stream, INTEGER_T_FORMAT, *(integer_t*)x);
    break;
  }
  case boolean_scheme_type:{
    if(*(integer_t*)(x)) fprintf(stream, "#t"); else fprintf(stream, "#f");
    break;
  }
  case character_scheme_type:{
    fprintf(stream, "%c", (char)*(integer_t*)x);
    break;
  }
  case symbol_scheme_type:{
    fprintf(stream, "%s", *(const char**)x);
    break;
  }
  case string_scheme_type:{
    fprintf(stream, "\"%s\"", (char*)x);
    break;
  }
  case cons_scheme_type:{
    fprintf(stream, "(");
    print_scheme_object(((word_t*)x)[0], stream);
    fprintf(stream, " . ");
    print_scheme_object(((word_t*)x)[1], stream);
    fprintf(stream, ")");
    break;
  }
  case closure_scheme_type:{
    fprintf(stream, "#<closure>");
    break;
  }
  case primitive_scheme_type:{
    fprintf(stream, "#<primitive>");
    break;
  }
  case variable_scheme_type:{
    fprintf(stream, "#<variable>");
    break;
  }
  case array_scheme_type:{
    fprintf(stream, "#(");
    const integer_t size = ((integer_t*)x)[0];
    integer_t i;
    for(i = 1; i <= size; i++){
      print_scheme_object(((word_t*)x)[i], stream);
      if(i != size)
        fprintf(stream, " ");
    }
    fprintf(stream, ")");
    break;
  }
  default:
    assert(0);
  } // switch  
}

void actual_main(void){
  printf("Initializing\n");
  eamgc_register_root_tracing_function(eamgc_trace_c_context_roots);
  eamgc_create_generic_pools();
  create_pools_for_scheme_objects();
  create_unique_scheme_objects();
  create_symbols_structures();
  printf("Starting\n");
  word_t list = scheme_empty_list;
  integer_t i = 0;
  while(1){
    //printf("%p\n", *(word_t*)make_symbol("la-mia-fava"));
    //list = make_cons(make_number(i), list);
    list = make_cons(make_symbol("abcaa"), list);
    i++;

    //    if((rand() % 1000) == 0)
    {
      print_scheme_object(list, stdout);
      printf("\n");
    }
  } // while
}
