/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include <stdio.h>
#include <math.h>

#include "gc.h"
#include "compile_time_parameters.h"
#include "roots.h"
#include "homogeneous.h"
#include "large.h"
#include "trace.h"
#include "roots.h"
#include "statistics.h"
#include "heuristics.h"
#include "../../common/set.h"
#include "../../common/doubly_linked_list.h"
#include "../../common/c_context.h"
#include "../../common/benchmark.h"


/* Was the GC initialized? */
static integer_t was_the_gc_initialized = 0u;

/* Support for GC disabling: */
static integer_t is_garbage_collector_enabled_ = (integer_t)1;
integer_t eamgc_is_garbage_collector_enabled(void){
  return is_garbage_collector_enabled_;
}
void eamgc_enable_garbage_collector(void){
  is_garbage_collector_enabled_ = (integer_t)1;
}
void eamgc_disable_garbage_collector(void){
  is_garbage_collector_enabled_ = (integer_t)0;
}

/* The total time spent for GC during the whole program run, in seconds: */
static double total_garbage_collection_time = 0.0;
static clock_t program_initialization_time;
static integer_t garbage_collections_no = 0;
static void print_final_garbage_collection_times(void){
  const clock_t program_finalization_time = clock();
  const double program_run_time = 
    (double)(program_finalization_time - program_initialization_time) /
    CLOCKS_PER_SEC;
  fprintf(stderr, "Collections performed: %i (%.3g secs per collection).\n",
          (int)garbage_collections_no,
          (garbage_collections_no != 0) ?
          total_garbage_collection_time / garbage_collections_no :
          0);
  fprintf(stderr, "Total GC time: %g sec (%.3g%%).\n",
          total_garbage_collection_time,
          (100.0 * total_garbage_collection_time / program_run_time));
}

/* Initialize the garbage collector and return. Note that this only works
   if the stack bottom is saved (see c_context.h), either before or
   after calling this function: */
void eamgc_initialize_garbage_collector(void){
  /* Refuse to initialize twice: */
  assert(! was_the_gc_initialized);

  /* Remember that we actually initialized the collector: */
  was_the_gc_initialized = 1;

  /* Create tracing stacks: */
  eamgc_initialize_trace_structures();
  
  /* Create needed data structures for the heap: */
  eamgc_initialize_homogeneous_objects_structures();
  eamgc_initialize_large_objects_structures();

  /* Create needed data structures for root tracing: */
  eamgc_initialize_roots_structures();

  /* Save the initialization time, and use it to print statistics at exit
     time: */
  program_initialization_time = clock();
  atexit(print_final_garbage_collection_times);
}

/* Initialize the garbage collector, and execute the given function. The
   collector will not work after the given gunction returns. 
   The function returns the value returned by continuation(): */
void eamgc_initialize_garbage_collector_and_continue(void (*function)(void)){
  /* Initialize everything except the stack bottom pointer: */
  eamgc_initialize_garbage_collector();
  
  /* Save the C stack bottom, so that we can find where it begins at
     root tracing time: */
  eam_save_stack_bottom_and_continue(function);
}

/* Perform a collection immediately: */
void eamgc_garbage_collect(void){
  double time_for_this_collection;

  EAM_START_BENCHMARK();
#ifdef ENABLE_ASSERTIONS
  fprintf(stderr, "\n[Collecting...\n");
  eamgc_dump_heap_size();
#endif
  
  /* A new collection is beginning. Set the alive heap size to zero. It
     will be incremented while finding alive objects in the marking phase: */
  eamgc_zero_alive_heap_size();
  
  /* Unmark everything: */
#ifdef ENABLE_ASSERTIONS
  EAM_START_BENCHMARK();
  fprintf(stderr, " * unmarking all objects from possibly nonfull homogeneous pages... ");
#endif
  eamgc_call_for_every_possibly_nonfull_homogeneous_page(eamgc_unmark_all_objects_in_homogeneous_page);
#ifdef ENABLE_ASSERTIONS
  EAM_STOP_BENCHMARK_AND_PRINT("done");
#endif

#ifdef ENABLE_ASSERTIONS
  /* This had a bug before: let's be defensive: */
  EAM_START_BENCHMARK();
  fprintf(stderr, " * checking that it worked... ");
  const struct eam_doubly_linked_list_element *p;
  for(p = eam_first_element(eamgc_get_all_homogeneous_pools()); p != NULL; p = eam_next(p))
    assert(eamgc_are_all_objects_unmarked_in_homogeneous_pool(
             (struct eamgc_homogeneous_pool*)eam_datum(p)));
  EAM_STOP_BENCHMARK_AND_PRINT("done");
#endif

  /* Trace roots: */
#ifdef ENABLE_ASSERTIONS
  EAM_START_BENCHMARK();
  fprintf(stderr, " * tracing roots... ");
#endif
  eamgc_trace_roots();
#ifdef ENABLE_ASSERTIONS
  EAM_STOP_BENCHMARK_AND_PRINT("done");
#endif
  
  /* Trace everything else: */
#ifdef ENABLE_ASSERTIONS
  EAM_START_BENCHMARK();
  fprintf(stderr, " * tracing heap... ");
#endif
  eamgc_trace();
#ifdef ENABLE_ASSERTIONS
  EAM_STOP_BENCHMARK_AND_PRINT("done");
#endif
  
  /* Reset all pools: */
#ifdef ENABLE_ASSERTIONS
  EAM_START_BENCHMARK();
  fprintf(stderr, " * resetting homogeneous pools... ");
#endif
  eamgc_call_for_every_homogeneous_pool(eamgc_reset_homogeneous_pool);

#ifdef ENABLE_ASSERTIONS
  EAM_STOP_BENCHMARK_AND_PRINT("done");
#endif

#ifdef ENABLE_ASSERTIONS
  /* Overwrite all dead objects, so that when there's a bug in the collector we
     notice it as soon as possible: */
  EAM_START_BENCHMARK();
  fprintf(stderr, " * overwriting dead homogenous objects... ");
  eamgc_call_for_every_homogeneous_page(eamgc_overwrite_dead_objects_in_homogeneous_page);
  EAM_STOP_BENCHMARK_AND_PRINT("done");
#endif

#ifdef ENABLE_ASSERTIONS
  EAM_START_BENCHMARK();
  fprintf(stderr, " * sweeping large objects... ");
#endif
  eamgc_sweep_large_objects();
#ifdef ENABLE_ASSERTIONS
  EAM_STOP_BENCHMARK_AND_PRINT("done");
#endif

#ifdef ENABLE_ASSERTIONS
  fprintf(stderr, "At the end of the collection:\n");
  eamgc_dump_heap_size();
  EAM_STOP_BENCHMARK_PRINT_AND_ASSIGN("done collecting]", time_for_this_collection);
  fprintf(stderr, "\n\n");
#else
  EAM_STOP_BENCHMARK_AND_ASSIGN(time_for_this_collection);
#endif
  
  /* Update GC statistics, so that we can base heuristics on the current
     situation: */
  eamgc_recompute_statistics();
  
  /* Update the number and the length of collections: */
  total_garbage_collection_time += (double)time_for_this_collection;
  garbage_collections_no ++;
}

void eamgc_finalize_garbage_collector(void){
  /* Refuse to finalize when the library is not initialized: */
  assert(was_the_gc_initialized);

  /* Remember that we actually finalized the collector: */
  was_the_gc_initialized = 0;
  
  /* Destroy tracing stacks: */
  eamgc_finalize_trace_structures();
  
  /* Destroy root tracing data structures: */
  eamgc_finalize_roots_structures();

  /* Destroy heap data structures: */
  eamgc_finalize_homogeneous_objects_structures();
  eamgc_finalize_large_objects_structures();
}

/* Return the tag associated to x, which must be a heap object (either
   homogeneous or large): */
unsigned_integer_t eamgc_object_to_tag(const word_t x){
  const eamgc_homogeneous_page_t homogeneous_page =
    eamgc_object_to_actual_homogeneous_page(x);
  if(homogeneous_page != NULL)
    return eamgc_homogeneous_page_to_tag(homogeneous_page);
  else{
#ifdef ENABLE_ASSERTIONS
    assert(eamgc_is_an_actual_large_object(x));
#endif // #ifdef ENABLE_ASSERTIONS
    return eamgc_large_object_to_tag(x);
  } // else
}

/* Return the data associated to x, which must be a heap object (either
   homogeneous or large): */
word_t eamgc_object_to_data(const word_t x){
  const eamgc_homogeneous_page_t homogeneous_page =
    eamgc_object_to_actual_homogeneous_page(x);
  if(homogeneous_page != NULL)
    return eamgc_homogeneous_page_to_data(homogeneous_page);
  else{
#ifdef ENABLE_ASSERTIONS
    assert(eamgc_is_an_actual_large_object(x));
#endif // #ifdef ENABLE_ASSERTIONS
    return eamgc_large_object_to_data(x);
  } // else
}
