/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "generic_pools.h"
#include "compile_time_parameters.h"

#include <string.h>


/* The array of generic pools, indexed by size: */
static struct eamgc_homogeneous_pool *generic_pools[EAMGC_MAXIMUM_GENERIC_OBJECT_SIZE + 1];

/* The only large pool we need: */
static struct eamgc_large_pool *generic_large_pool;

static integer_t were_generic_pools_created = 0;

void eamgc_create_generic_pools(void){
#ifdef ENABLE_ASSERTIONS
  assert(! were_generic_pools_created);
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* Create pools for all small sizes: */
  integer_t i;
  for(i = 1; i <= EAMGC_MAXIMUM_EXACTLY_SUPPORTED_GENERIC_OBJECT_SIZE; i++)
    generic_pools[i] =
      eamgc_create_homogeneous_pool((unsigned_integer_t)i,
                                 0u,
                                 NULL,
                                 eamgc_homogeneous_trace_everything_function);
  
  /* From now on we create pools for size i, 2i, 3i... until
     EAMGC_MAXIMUM_GENERIC_OBJECT_SIZE. We share the same pool for all the
     sizes in (k i + 1, (k + 1)i ], for every k: */
  integer_t size;
  for(size = EAMGC_MAXIMUM_EXACTLY_SUPPORTED_GENERIC_OBJECT_SIZE * 2;
      size <= EAMGC_MAXIMUM_GENERIC_OBJECT_SIZE;
      size *= 2){
    struct eamgc_homogeneous_pool *pool =
      eamgc_create_homogeneous_pool(size,
                                 0u,
                                 NULL,
                                 eamgc_homogeneous_trace_everything_function);
    for(i = size; i > (size / 2); i--){
#ifdef ENABLE_ASSERTIONS
      assert(i <= EAMGC_MAXIMUM_GENERIC_OBJECT_SIZE);
      assert(i >= 0);
      assert(i > EAMGC_MAXIMUM_EXACTLY_SUPPORTED_GENERIC_OBJECT_SIZE);
#endif 
      generic_pools[i] = pool;
    } // inner for
  } // outer for
#ifdef ENABLE_ASSERTIONS
  assert(size == EAMGC_MAXIMUM_GENERIC_OBJECT_SIZE * 2);
#endif
  
  /* Create the large pool: */
  generic_large_pool = eamgc_create_large_pool(0u,
                                            NULL,
                                            eamgc_large_trace_everything_function,
                                            NULL);
  /* Remember we created generic pools: */
  were_generic_pools_created = 1;
}

/* Finalize generic pools support: */
void eamgc_destroy_generic_pools(void){
#ifdef ENABLE_ASSERTIONS
  assert(were_generic_pools_created);
#endif // #ifdef ENABLE_ASSERTIONS
  // To do: enable pool destruction.
  were_generic_pools_created = 0;
}

/* Finalize generic pools support only if it was initialized before: */
void eamgc_destroy_generic_pools_if_needed(void){
  if(were_generic_pools_created)
    eamgc_destroy_generic_pools();
}

/* Return the generic homogeneous pool from which to allocate an object of 
   the given size. The size is in words. NULL is returned if no homogeneous
   pool supports large enough objects: */
static struct eamgc_homogeneous_pool* eamgc_size_to_generic_homogeneous_pool(
   const unsigned_integer_t size)
  __attribute__ (( pure ));
static struct eamgc_homogeneous_pool* eamgc_size_to_generic_homogeneous_pool(const unsigned_integer_t size){
#ifdef ENABLE_ASSERTIONS
  /* Size must be strictly greater than 0: */
  assert(size > 0);
#endif
  if(size > EAMGC_MAXIMUM_GENERIC_OBJECT_SIZE)
    /* We have no homogeneous pools for objects larger than
       EAMGC_MAXIMUM_GENERIC_OBJECT_SIZE words: */
    return NULL;
  else{
    #ifdef VERBOSE_DEBUG
    //    fprintf(stderr, "eamgc_size_to_generic_homogeneous_pool(): Size: %u -> %u\n",
    //            (unsigned)size,
    //            (unsigned)generic_pools[size]->object_size);
    #endif // #ifdef VERBOSE_DEBUG
#ifdef ENABLE_ASSERTIONS
    /*
    fprintf(stderr, "%i should-be-\">\"-than %i\n",
            (int)generic_pools[size]->object_size,
            (int)size);
    */
    assert(eamgc_homogeneous_pool_to_object_size(generic_pools[size]) >= size);
#endif
    /* just a fast array access: */
    return generic_pools[size];
  } // else
}

/* Allocate an object from the most appropriate generic pool. Note that 
   the size is in words, as usual; hence this interface is *not* compatible
   with malloc(). */
word_t eamgc_allocate_from_generic_pool(const unsigned_integer_t size){
  /* Look for the most appropriate pool: */
  struct eamgc_homogeneous_pool *pool = eamgc_size_to_generic_homogeneous_pool(size);
  
  /* Allocate from the homogeneous pool we found, if any; otherwise allocate
     a large object: */
  if(pool != NULL)
    return eamgc_allocate_from_homogeneous_pool(pool);
  else
    return eamgc_allocate_from_large_pool(generic_large_pool, size);
}

void* eamgc_malloc(size_t size_in_bytes){
  /* Here the size is in bytes, but eamgc_allocate_from_generic_pool() needs
     it in words: */
  unsigned_integer_t size_in_words = size_in_bytes / sizeof(word_t);
  if((size_in_bytes % sizeof(word_t)) != 0)
    size_in_words++;
#ifdef ENABLE_ASSERTIONS
  assert((size_in_words * sizeof(word_t)) >= size_in_bytes);
#endif
  return (void*)eamgc_allocate_from_generic_pool(size_in_words);
}

void* eamgc_calloc(size_t count, size_t element_size_in_bytes){
  /* Create a block, clear it and return it: */
  const size_t size_in_bytes = count * element_size_in_bytes;
  void *r = eamgc_malloc(size_in_bytes);
  memset(r, 0, size_in_bytes);
  return r;
}

void* eamgc_realloc(void *p, size_t new_size_in_bytes){
  /* If p is NULL a realloc()-compatible function should return the same
     as malloc(), i.e. the object we've just allocated: */
  if(p == NULL)
    return eamgc_malloc(new_size_in_bytes);
  
  /* We need to know the objects' sizes, even better if in bytes: */
  const unsigned_integer_t old_size_in_words =
    eamgc_is_an_actual_large_object((word_t)p) ?
    eamgc_large_object_to_size((word_t)p) :
    eamgc_homogeneous_object_to_size((word_t)p);
  const unsigned_integer_t old_size_in_bytes =
    old_size_in_words * sizeof(word_t);
  
  /* Just return p unchanged if the new size is the same as the current
     one: */
  if(old_size_in_bytes == new_size_in_bytes)
    return p;
  
  /* We actually have to copy something: see how many bytes we have to
     copy: it's the minimum between the two sizes:*/
  const unsigned_integer_t how_many_bytes_to_copy =
    (old_size_in_bytes >= new_size_in_bytes) ?
    new_size_in_bytes :
    old_size_in_bytes;
  
  /* Make a new object, fill it and return it: */
  void *r = eamgc_malloc(new_size_in_bytes);
  memcpy(r, p, (size_t)how_many_bytes_to_copy);
  return r;
}
