/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EAMGC_GENERIC_POOLS_H_
#define EAMGC_GENERIC_POOLS_H_

#include "gc.h"

/* "Generic" pools are pools with conservative pointer finding and no 
   finalization support, allowing the user to request an allocation of
   some size with may not be known at compile time. The "most appopriate"
   pool is chosen, i.e. the one for objects of the minimum size greater
   than or equal to the requested one. If no such pool exists among
   generic pools then a large object is allocated. */


/* Create generic pools of various sizes: */
void eamgc_create_generic_pools(void);

/* Finalize generic pools support: */
void eamgc_destroy_generic_pools(void);

/* Finalize generic pools support only if it was initialized before: */
void eamgc_destroy_generic_pools_if_needed(void);

/* Allocate an object from the most appropriate generic pool. Note that 
   the size is in words, as usual; hence this interface is *not* compatible
   with malloc(). */
word_t eamgc_allocate_from_generic_pool(const unsigned_integer_t size);

/* Compatible wrappers emulating libc functions, of course with GC 
   functionality. Any failure is fatal, so malloc() and friends never
   return NULL. Here the size is in bytes, for compatibility. */
void* eamgc_malloc(size_t size_in_bytes) __attribute__ (( malloc ));
void* eamgc_calloc(size_t elements_no, size_t element_size_in_bytes)
  __attribute__ (( malloc ));

/* Do absolutely nothing, except evaluating the parameter, which may have side
   effects. Using a macro instead of a function allows the optimizer to entirely
   remove the call in most cases, even with separate compilation: */
#define eamgc_free(p) \
  { const void *__useless __attribute__ (( unused )) = (const void*)(p); }

/* Just make a new (possibly truncated) copy of the block. The old block
   will be collected when appropriate. The size is in bytes, for
   compatibility. */
void* eamgc_realloc(void *p, size_t new_size_in_bytes)
  __attribute__ (( malloc ));

#endif // #ifndef EAMGC_GENERIC_POOLS_H_
