/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2003, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with epsilon; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */


//#define EAMGC_BITS_FIELD // it's faster without this

#include "../../common/types.h"
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <limits.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "gc.h"
#include "compile_time_parameters.h"
#include "homogeneous.h"
#include "statistics.h"
#include "heuristics.h"
#include "../../common/malloc.h"
#include "../../common/set.h"
#include "../../common/list.h"
#include "../../common/command_line.h"
#include "../../common/doubly_linked_list.h"

/* The implementation of each homogeneous page pool: */
struct eamgc_homogeneous_pool{
  /* How large a single object is, in words: */
  unsigned_integer_t object_size;
  
  /* The tag shared by all objects in this pool: */
  unsigned_integer_t tag;
  
  /* Generic data which the user can associate to the whole pool. It's the
     user responsibility to allocate and free them, if ever needed. */
  word_t data;
  
  /* The number of *objects* (not words) fitting in a page of this pool: */
  integer_t objects_per_page;

  /* The procedure used to mark the objects referred by an object of this
     pool: this must not be NULL. */
  eamgc_homogeneous_trace_function_t trace_function;
  
  /* The page we're currently using for allocation. This must never be NULL
     except just before collecting, when no possibly nonfull pages are left
     *and* we are about to trigger a collection (which always fixes this 
     anomaly). We don't want to check for presence at every allocation; we 
     don't even want to access the page from a list at every allocation, so
     we keep the 'first' page here (and, by convention, we *don't* keep the
     same pointer in possibly_nonfull_pages): */
  eamgc_homogeneous_page_t current_page;
  
  /* List of other pages which may have some free space: */
  struct eam_doubly_linked_list* possibly_nonfull_pages;
  
  /* List of surely full pages: */
  struct eam_doubly_linked_list* full_pages;
}; // struct

/* The header is the first part of a page: in the same memalign()'d block there
   are the GC bits, the allocated bits and the payload (also pointed by the
   eamgc_bits, allocated_bits and payload fields of the header, for efficiency) */
struct eamgc_homogeneous_page_header{ /* Size must be a multiple of a word size */
  /* These fields are copied from the pool holding the page. It's nice to have
     them here, because access to them must be as fast as possible: */
  unsigned_integer_t object_size; /* in words */
  unsigned_integer_t tag;
  word_t data; /* data used for the whole pool */
  integer_t objects_per_page; /* how many object this page *can* hold */
  unsigned_integer_t alive_objects_no; /* alive objects no. in this page at last GC */
  eamgc_homogeneous_trace_function_t trace_function; /* never NULL: it can safely be called */
  
  /* Pointer to the array of GC bits (they are within the same memalign()ed
     page, after the header. Word-aligned): */
#ifdef EAMGC_BITS_FIELD
  unsigned_integer_t* eamgc_bits;
#endif
  
  /* Pointer to the payload (it's within the same memalign()ed page, after 
     the GC bits. Word-aligned): */
  const word_t* payload; /* pointer to the first word of the payload */
  
  /* Index of the next object in the page to be *considered* for allocation. 
     This means that if the referred object is marked then the following one
     will be considered, and so on. */
  integer_t allocation_index;
}; // struct


/* This works since homogeneous_page_header has a size which is multiple of a word */
#define HEADER_SIZE (sizeof(struct eamgc_homogeneous_page_header) / sizeof(word_t)) /* in words */

/* The number of bits in a word: */
#define WORD_BIT (CHAR_BIT * sizeof(word_t))

/* Returns the number of objects which fit in a page; OBJECT_SIZE is in
   words: */
#define OBJECTS_IN_A_HOMOGENEOUS_PAGE(OBJECT_SIZE) \
  (((integer_t) \
    (((float)(EAMGC_PAGE_SIZE) - (float)(HEADER_SIZE)) /              \
     ((float)(OBJECT_SIZE) + 1.0 /* a GC bit */ / (float)(WORD_BIT)))) - 1)

/* Returns the size in words of the GC bit-array for a page of objects
   with size OBJECT_SIZE (in words) */
#define EAMGC_BITS_SIZE(OBJECT_SIZE) \
  ((integer_t)ceil(((float)OBJECTS_IN_A_HOMOGENEOUS_PAGE(OBJECT_SIZE)) / (float)WORD_BIT))

/* Given the address of a page, return the address of the first word of
   its GC bits. GC bits start right after the header. Instead of this
   (admittedly ugly) solution we could have a eamgc_bits field in the header
   structure; this actually was the previous solution. But in this way we
   can save a load every time we access GC bits (not so important, since
   probably it'd be a cache hit), and I think there are more opportunities
   for optimization by GCC.
   - Benchmarked on 2006-05-25: the new solution is *much* better than the
   old one (GC time down by ~45% on my machine). Try #defin'ing EAMGC_BITS_FIELD
   at the beginning of the file to compare. */
#ifdef EAMGC_BITS_FIELD
#define PAGE_TO_GC_BITS(page) \
  ((page)->eamgc_bits)
#else
#define PAGE_TO_GC_BITS(page) \
  ((unsigned_integer_t*)(((char*)(page)) + sizeof(struct eamgc_homogeneous_page_header)))
#endif // #ifdef EAMGC_BITS_FIELD

/* The current homogeneous heap_size, payload only: we exclude headers,
   GC bits and all data structures external to pages. */
static integer_t homogeneous_heap_size = 0;

/* How many pages currently exist: */
static integer_t homogeneous_pages_no = 0;
integer_t eamgc_get_homogeneous_pages_no(void){
  return homogeneous_pages_no;
}

/* Return the total memory space (in words) used by homogeneous page payloads.
   Headers, GC bits and other data structures are not counted: */
unsigned_integer_t eamgc_get_homogeneous_heap_size(void){
  return (unsigned_integer_t)homogeneous_heap_size;
}

/* Given an object size, return the number of objects of that size fitting in
   a single homogeneous page: */
integer_t eamgc_homogeneous_object_size_to_object_no(const unsigned_integer_t object_size){
#ifdef ENABLE_ASSERTIONS
  assert(OBJECTS_IN_A_HOMOGENEOUS_PAGE(object_size) > 0);
#endif
  return OBJECTS_IN_A_HOMOGENEOUS_PAGE(object_size);
}

/* The associative data structure set used to quickly find whether a given
   homogeneous page exists: */
static set_t set_of_homogeneous_pages;
void eamgc_initialize_homogeneous_objects_structures(void){
  /* Create the set: */
  set_of_homogeneous_pages =
    eam_create_set(EAMGC_SIZE_OF_SET_OF_HOMOGENEOUS_PAGES);

  /* Reset counters: */
  homogeneous_heap_size = 0;
  homogeneous_pages_no = 0;
}

/* The linked list of all pools, useful for cleaning up at GC termination time.
   Note that NULL means uninitialized, and it's not an empty list. */
static struct eam_doubly_linked_list *pool_list = NULL;

/* Return a pointer to a static doubly-linked list containing all the
   homogeneous pools. The user should not modify the returned object in
   any way: */
struct eam_doubly_linked_list* eamgc_get_all_homogeneous_pools(void){
  return pool_list;
}

/* After creating a pool you should always register it with this function, so
   that things get cleaned up at program termination. Each pool must be
   registered exactly once: */
static void register_homogeneous_pool(struct eamgc_homogeneous_pool *pool){
  /* Initialize the global pool list, if this was not already done: */
  if(pool_list == NULL)
    pool_list = eam_create_doubly_linked_list();
  
  /* Add the new pool to the current list of pools: */
  eam_prepend_to_doubly_linked_list(pool, pool_list);
}

/* *Not* for the user! This assumes some data structures have been already
   destroyed. */
static void destroy_homogeneous_pool(struct eamgc_homogeneous_pool *p){
  /* Destroy what remains of the lists of pages (elements should have 
     already been destroyed): */
  eam_destroy_doubly_linked_list_structure(p->possibly_nonfull_pages);
  eam_destroy_doubly_linked_list_structure(p->full_pages);
  
  /* Destroy the buffer holding together the pool: */
  free(p);
}

/* Finalize internal data structures needed to manage homogeneous pools. The
   user never needs to directly call this. */
void eamgc_finalize_homogeneous_objects_structures(void){
  /* Destroy the hash: */
  eam_destroy_set(set_of_homogeneous_pages);
  
  /* Destroy pages: */
  eamgc_call_for_every_homogeneous_page(eamgc_destroy_homogeneous_page);

  /* Destroy pools: */
  eamgc_call_for_every_homogeneous_pool(destroy_homogeneous_pool);
  
  /* Reset the list of pools to its 'uninitialized' state: */
  eam_destroy_doubly_linked_list_structure(pool_list);
  pool_list = NULL;
}

static inline word_t index_to_object_in_homogeneous_page(const integer_t index,
                                                         const eamgc_homogeneous_page_t p)
  __attribute__ (( pure ));

static void clear_eamgc_bits_in_homogeneous_page(eamgc_homogeneous_page_t page);

eamgc_homogeneous_page_t eamgc_create_homogeneous_page(struct eamgc_homogeneous_pool *pool){  
#ifdef VERBOSE_DEBUG
  fprintf(stderr, "[+ creating a new homogeneous page +]\n");
#endif // #ifdef VERBOSE_DEBUG
  /* Allocate storage (header + payload) */
  eamgc_homogeneous_page_t r = (eamgc_homogeneous_page_t)
    memalign(EAMGC_PAGE_SIZE_IN_CHARS, EAMGC_PAGE_SIZE_IN_CHARS);
  
  /* Fail if memalign() failed: */
  if(r == NULL)
    eam_fatal("could not allocate homogeneous page");

  /* Copy some data from the pool. We want to have them also here because
     it's important they're fast to access: */
  r->object_size = pool->object_size;
  r->tag = pool->tag;
  r->data = pool->data;
  r->objects_per_page = pool->objects_per_page;
  r->trace_function = pool->trace_function;
  
  /* Initialize other fields: */
  r->alive_objects_no = 0u;
  r->allocation_index = 0;
#ifdef EAMGC_BITS_FIELD
  r->eamgc_bits = (unsigned_integer_t*)((char*)r + sizeof(struct eamgc_homogeneous_page_header));
#endif
  r->payload = ((word_t*)((char*)r +
			  sizeof(struct eamgc_homogeneous_page_header))) +
                          EAMGC_BITS_SIZE(pool->object_size);
  
  /* Start with the GC bits cleared: */
  clear_eamgc_bits_in_homogeneous_page(r);
  
  /* Insert into the set of homogeneous pages: */
  eam_insert_pointer_into_set((word_t)r, set_of_homogeneous_pages);
  
  /* We have one more page now: */
  homogeneous_pages_no ++;
  homogeneous_heap_size +=
    (integer_t)r->objects_per_page * (integer_t)r->object_size;

#ifdef ENABLE_ASSERTIONS
  /* Cover all objects words with the "uninitialized" pattern: */
  integer_t i;
  for(i = 0; i < r->objects_per_page; i++)
    eamgc_overwrite_homogeneous_object(index_to_object_in_homogeneous_page(i, r),
                                    eamgc_uninitialized_pattern);
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* This shouldn't be needed, but just to be sure: */
#ifdef ENABLE_ASSERTIONS
  assert(r->alive_objects_no == 0u);
  assert(r->allocation_index == 0);
#endif // #ifdef ENABLE_ASSERTIONS

  /* Return: */
  return r;
}

/* Free the storage for the homogeneous page p (if possible on this system),
   and de-register p as a homogeneous page. */
void eamgc_destroy_homogeneous_page(eamgc_homogeneous_page_t p){
  /* Remove the page from the set of homogeneous pages: */
  eam_remove_pointer_from_set((word_t)p, set_of_homogeneous_pages);
  
  /* We're destroying one page: */
  homogeneous_pages_no --;
  homogeneous_heap_size -=
    (integer_t)p->objects_per_page * (integer_t)p->object_size;
  
#ifdef CAN_FREE_MEMALIGNED_BLOCKS
  /* Free storage: */
  free(p); /* free()'ing memalign()'ed blocks is only safe on GNU systems! */
#endif
}

/* Clear all GC bits of p, which must be an existing homogeneous page. */
static void clear_eamgc_bits_in_homogeneous_page(eamgc_homogeneous_page_t page){
  /* Clear GC bits (not bit by bit, of course), and not even word by word
     if possible... This is crucial for efficiency. */
  memset(PAGE_TO_GC_BITS(page), 
         0, 
         (char*)(page->payload) - (char*)(PAGE_TO_GC_BITS(page)));
#ifdef ENABLE_ASSERTIONS
  assert(eamgc_are_all_objects_unmarked_in_homogeneous_page(page));
#endif // #ifdef ENABLE_ASSERTIONS
}

/* Return the index of object in the payload of p, i.e. the index of its
   GC bit, starting from 0. Note that this also word for *internal* pointers,
   at no extra cost: */
static inline integer_t object_to_index_in_homogeneous_page(const word_t object,
                                                     const eamgc_homogeneous_page_t p)
  __attribute__ (( pure ));
static inline integer_t object_to_index_in_homogeneous_page(const word_t object,
                                                     const eamgc_homogeneous_page_t p){
  return (integer_t)
    (((word_t*)object - (word_t*)(p->payload)) / (p->object_size));
}

/* Convert a possibly internal pointer to an existing homogeneous object into
   a pointer to the first word of the same homogeneous object:  */
word_t eamgc_possibly_internal_pointer_to_homogeneous_object(
    const word_t possibly_internal_pointer){
  /* Extract the page from the internal pointer: */
  const eamgc_homogeneous_page_t page = 
    eamgc_object_to_hypothetical_homogeneous_page(possibly_internal_pointer);
  
  /* Get the index of the object (this also works with internal pointers): */
  const integer_t index = 
    object_to_index_in_homogeneous_page(possibly_internal_pointer, page);
  
  /* Convert back from index to object: */
  return index_to_object_in_homogeneous_page(index, page);
}

// To do: remove this kludge
integer_t eamgc_object_to_index_in_homogeneous_page(const word_t object,
                                                 const eamgc_homogeneous_page_t p){
  return object_to_index_in_homogeneous_page(object, p);
}

/* The index must be in the correct range: */
static inline word_t index_to_object_in_homogeneous_page(const integer_t index,
                                                         const eamgc_homogeneous_page_t p){
#ifdef ENABLE_ASSERTIONS
  assert(index >= 0);
  assert(index < p->objects_per_page);
#endif // #ifdef ENABLE_ASSERTIONS
  return (word_t)(((word_t*)p->payload) + index * (integer_t)(p->object_size));
}


/* object must belong to p. Returns the GC bit as it was before: */
integer_t eamgc_mark_in_homogeneous_page(const word_t object,
                                         const eamgc_homogeneous_page_t p){
  /* Find the right GC bit: */
  const integer_t index_of_object = 
    object_to_index_in_homogeneous_page(object, p);
  
  unsigned_integer_t* word_of_eamgc_bits =
    &(PAGE_TO_GC_BITS(p)[index_of_object / WORD_BIT]);
  const unsigned_integer_t bitmask = 
    (unsigned_integer_t)1u << (index_of_object % WORD_BIT);
  
  /* If the object was already marked then return 1 and do nothing: */
  if(*word_of_eamgc_bits & bitmask)
    return 1;
  
  /* The object was not marked. Mark it, setting its GC bit: */
  *word_of_eamgc_bits |= bitmask;
  
  /* We marked a previously unmarked object. In other words we found an alive
     object: */
  eamgc_increment_alive_homogeneous_heap_size(p->object_size);
  p->alive_objects_no++;
  
  /* If we arrived here then the object was not previously marked: */
  return 0;
}

/* object must belong to p */
inline void eamgc_unmark_in_homogeneous_page(const word_t object,
                                             const eamgc_homogeneous_page_t p){
  /* Find the right GC bit: */
  const integer_t index_of_object = object_to_index_in_homogeneous_page(object, p);
  
  /* Set the bit: */
  PAGE_TO_GC_BITS(p)[index_of_object / WORD_BIT] &= 
    ~ ((unsigned_integer_t)1u << (index_of_object % WORD_BIT));
}

/* The index must be in the correct range: */
inline void unmark_index_in_homogeneous_page(integer_t index,
                                             eamgc_homogeneous_page_t p){
  /* Just set the bit: */
  PAGE_TO_GC_BITS(p)[index / WORD_BIT] &=
    ~ ((unsigned_integer_t)1u << (index % WORD_BIT));
}

/* object must belong to p */
inline integer_t eamgc_is_marked_in_homogeneous_page(word_t object, 
                                                  eamgc_homogeneous_page_t p){
#ifdef ENABLE_ASSERTIONS
  assert(eamgc_object_to_actual_homogeneous_page(object) == p);
#endif
  
  /* Find the right GC bit: */
  const integer_t index_of_object = object_to_index_in_homogeneous_page(object, p);

  /* Return nonzero if the bit is set: */
  return PAGE_TO_GC_BITS(p)[index_of_object / WORD_BIT] 
       & ((unsigned_integer_t)1u << (index_of_object % WORD_BIT));
}

/* The index must be in the right range */
inline integer_t is_index_marked_in_homogeneous_page(const integer_t index,
                                                     const eamgc_homogeneous_page_t p)
  __attribute__ (( pure ));
inline integer_t is_index_marked_in_homogeneous_page(const integer_t index,
                                                     const eamgc_homogeneous_page_t p){
#ifdef ENABLE_ASSERTIONS
  assert(index >= 0);
  assert(index < p->objects_per_page);
#endif
  
  /* Return nonzero if the bit is set: */
  return PAGE_TO_GC_BITS(p)[index / WORD_BIT] & 
    ((unsigned_integer_t)1u << (index % WORD_BIT));
}

#define FROM_0_TO(n) (((integer_t)((double)rand()/(double)RAND_MAX * (double)(n))))
#define FROM_TO(a, b) (a + FROM_0_TO(b - a + 1))

inline integer_t eamgc_is_homogeneous_page_full(const eamgc_homogeneous_page_t p){
#ifdef ENABLE_ASSERTIONS
  assert(p->allocation_index >= 0);
  assert(p->allocation_index <= p->objects_per_page);
#endif // #ifdef ENABLE_ASSERTIONS
  return p->allocation_index == p->objects_per_page;
}

const word_t eamgc_uninitialized_pattern = (word_t)0xaaaaaaaa;
const word_t eamgc_just_created_pattern = (word_t)0xffffffff;
const word_t eamgc_dead_pattern = (word_t)0xdeaddead;

void eamgc_overwrite_homogeneous_object(const word_t object, const word_t pattern){
  word_t *object_as_word_star = (word_t*)object;
  const eamgc_homogeneous_page_t page = eamgc_object_to_actual_homogeneous_page(object);
  integer_t i;
  for(i = 0; i < page->object_size; i++)
    object_as_word_star[i] = pattern;
}

/* Overwrite all unmarked objects in the page with the 'dead' pattern. Useful 
   for debugging. Of course this should be called only after a complete
   marking. */
void eamgc_overwrite_dead_objects_in_homogeneous_page(const eamgc_homogeneous_page_t p){
  integer_t i;
  const integer_t objects_no = p->objects_per_page;
  for(i = 0; i < objects_no; i++)
    if(! is_index_marked_in_homogeneous_page(i, p)){
      word_t object = index_to_object_in_homogeneous_page(i, p);
#ifdef VERBOSE_DEBUG
      if((((word_t*)object)[0] != eamgc_dead_pattern) &&
         ((word_t*)object)[0] != eamgc_uninitialized_pattern){
        fprintf(stderr, "!!!!! DEAD object: ");
        eamgc_dump_homogeneous_object(object, stderr);
        fprintf(stderr, "\n");
      } // inner if
#endif // #ifdef VERBOSE_DEBUG
      eamgc_overwrite_homogeneous_object(object, eamgc_dead_pattern);
    } // if
}

void eamgc_dump_homogeneous_object(const word_t object, FILE *file){
  /* Check that object is actually a homogeneous object: */
  assert(eamgc_object_to_actual_homogeneous_page(object) != NULL);
  
  /* Write an open bracket, to mark the beginning of the object: */
  fprintf(file, UNSIGNED_INTEGER_T_FORMAT":[ ", (unsigned_integer_t)object);

  /* Write each word: */
  const word_t *object_as_word_star = (word_t*)object;
  const unsigned_integer_t size = eamgc_homogeneous_object_to_size(object);
  unsigned_integer_t i;
  for(i = 0u; i < size; i++)
    fprintf(file, 
            UNSIGNED_INTEGER_T_FORMAT" ",//HEXADECIMAL_UNSIGNED_INTEGER_T_FORMAT" ",
            ((unsigned_integer_t)*(object_as_word_star + i)));
  /* Write a closed bracket, to mark the beginning of the object: */
  fprintf(file, "]");
}

/* Returns NULL on failure: */
word_t eamgc_allocate_from_homogeneous_page(const eamgc_homogeneous_page_t p){
  /* Look for a non-allocated object starting from the allocation index,
     until either we find a non-allocated object or we discover that the
     page is full (i.e. the allocation index is past the end of the
     page).
     Note that we don't use p->allocation_index as an interation variable
     because we want to save the cost of a load and a store per iteration.
     We work with pointers here (for GC bits), and the code is difficult 
     to optimize for the compiler. Let's help it: */
  integer_t candidate_index; // only *one* load and *one* store
  const integer_t objects_per_page = p->objects_per_page; // only *one* load
  for(candidate_index = p->allocation_index;
      candidate_index < objects_per_page;
      candidate_index++){
    
    /* Is the candidate object we're considering non-allocated? If so,
       just set the allocation index so that the next time we consider
       the *following* object, and return the object we found.
       This should be by far the most common case, and here it's important
       that GCC guesses branch prediction right. It's probably worth the
       trouble to compile this source file with profile-driven feedback
       optimizations. */
    if(! is_index_marked_in_homogeneous_page(candidate_index, p)){
      p->allocation_index = candidate_index + 1;
      const word_t object = 
        index_to_object_in_homogeneous_page(candidate_index, p);
#ifdef VERBOSE_DEBUG
      /* Very crude heuristic approach: */
      if((((word_t*)object)[0] != eamgc_dead_pattern) &&
         (((word_t*)object)[0] != eamgc_uninitialized_pattern)){
        fprintf(stderr, "? WARNING: allocating over ");
        eamgc_dump_homogeneous_object(object, stderr);
        fprintf(stderr, "\n");
      } // if
#endif // #ifdef VERBOSE_DEBUG

#ifdef ENABLE_ASSERTIONS
      eamgc_overwrite_homogeneous_object(object, eamgc_just_created_pattern);
      assert(! eamgc_is_marked_in_homogeneous_page(object, p));
#endif // #ifdef ENABLE_ASSERTIONS
      return object;
    } // if
    
    /* Unmark the bit we just read; the object is currently alive, but it
       may die before the next trace: */
    unmark_index_in_homogeneous_page(candidate_index, p);
    
    /* We weren't so lucky, the candidate object is already allocated.
       Try with the next one. */
  } // for
  
  /* If we arrived here then we didn't find a non-allocated object.
     Allocating from this page failed: remember it's full: */
#ifdef ENABLE_ASSERTIONS
  assert(candidate_index == objects_per_page);
#endif
  p->allocation_index = objects_per_page;
  return NULL;
}

/* This is only a computation on pointers; it does *not* check whether
   object actually belongs to a homogeneous page. */
inline eamgc_homogeneous_page_t eamgc_object_to_hypothetical_homogeneous_page(
    const word_t object){
  unsigned_integer_t page_as_number;

  /* Set to zero the EAMGC_PAGE_OFFSET_WIDTH least-significant bits of
     object: */
  page_as_number = (unsigned_integer_t)object;
  page_as_number >>= EAMGC_PAGE_OFFSET_WIDTH;
  page_as_number <<= EAMGC_PAGE_OFFSET_WIDTH;

  /* Return the number re-cast to a pointer: */
  return (eamgc_homogeneous_page_t)page_as_number;
}

/* Returns NULL if the object does not belong to an homogeneous page,
   else return its page: */
eamgc_homogeneous_page_t eamgc_object_to_actual_homogeneous_page(const word_t object){
  /* Set to zero the EAMGC_PAGE_OFFSET_WIDTH least-significant bits of
     object: */
  const unsigned_integer_t page_as_number = (unsigned_integer_t)
    eamgc_object_to_hypothetical_homogeneous_page(object);
  
  /* If the hypothetical page is not an actual homogeneous page then fail: */
  if(! eam_does_pointer_belong_to_set((word_t)page_as_number,
				  set_of_homogeneous_pages))
    return NULL;
  /* If the hypothetical address is not in the payload then fail: */
  if(object < (word_t)((const eamgc_homogeneous_page_t)page_as_number)->payload)
    return NULL;

  /* If the hypothetical address is an internal pointer than fail: */
  if(((char*)object - (char*)(((eamgc_homogeneous_page_t)page_as_number)->payload)) %
     (sizeof(word_t) * ((eamgc_homogeneous_page_t)page_as_number)->object_size) != 0){
#ifdef VERBOSE_DEBUG
    fprintf(stderr,"WARNING: %p is an internal (false) pointer\n", object);
#endif
    return NULL;
  }
  
  /* If we arrived here then object points to an allocated object of a
     homogeneous page: */
  return (eamgc_homogeneous_page_t)page_as_number;
}

/* Given an object from a homogeneous page (we don't actually check whether
   it's really a homogeneous page!) return its tag: */
unsigned_integer_t eamgc_homogeneous_object_to_tag(const word_t object){
  /* Compute a pointer to the object's page. This should be very fast: */
  const eamgc_homogeneous_page_t page = 
    eamgc_object_to_hypothetical_homogeneous_page(object);
#ifdef ENABLE_ASSERTIONS
  assert(page == eamgc_object_to_actual_homogeneous_page(object));
#endif
  
  /* Take what we want from the page header: */
  return page->tag;
}

/* Given an object from a homogeneous page (we don't actually check whether
   it's really a homogeneous page!) return its pool-relative data: */
word_t eamgc_homogeneous_object_to_data(const word_t object){
  /* Compute a pointer to the object's page. This should be very fast: */
  const eamgc_homogeneous_page_t page = 
    eamgc_object_to_hypothetical_homogeneous_page(object);
#ifdef ENABLE_ASSERTIONS
  assert(page == eamgc_object_to_actual_homogeneous_page(object));
#endif
  
  /* Take what we want from the page header: */
  return page->data;
}

/* Return the data of the given homogeneous pool: */
inline word_t eamgc_homogeneous_pool_to_data(const struct eamgc_homogeneous_pool *pool){
  return pool->data;
}

/* Return the object size of the given homogeneous pool: */
inline unsigned_integer_t eamgc_homogeneous_pool_to_object_size(const struct eamgc_homogeneous_pool *pool){
  return pool->object_size;
}

/* Given an object from a homogeneous page (we don't actually check whether
   it's really a homogeneous page!) return its allocated size in words: */
unsigned_integer_t eamgc_homogeneous_object_to_size(const word_t object){
  /* Compute a pointer to the object's page. This should be very fast: */
  const eamgc_homogeneous_page_t page = 
    eamgc_object_to_hypothetical_homogeneous_page(object);
#ifdef ENABLE_ASSERTIONS
  assert(page == eamgc_object_to_actual_homogeneous_page(object));
#endif
  
  /* Take what we want from the page header: */
  return page->object_size;
}

/* Given a homogeneous page (of course we don't actually check that
   it's really a homogeneous page) return the size (in words) of one
   of its objects: */
inline unsigned_integer_t eamgc_homogeneous_page_to_object_size(
    const eamgc_homogeneous_page_t page){
  return page->object_size;
}

/* Return the tag of the given homogeneous page: */
inline unsigned_integer_t eamgc_homogeneous_page_to_tag(const eamgc_homogeneous_page_t page){
  return page->tag;
}

/* Return the object data of the given homogeneous page: */
inline word_t eamgc_homogeneous_page_to_data(const eamgc_homogeneous_page_t page){
  return page->data;
}

/* Unmark all the objects in the homogeneous page. */
void eamgc_unmark_all_objects_in_homogeneous_page(const eamgc_homogeneous_page_t page){
  clear_eamgc_bits_in_homogeneous_page(page);
}

/* Return a nonzero value iff all the objects in the given homogeneous page
   are *not* marked: */
integer_t eamgc_are_all_objects_unmarked_in_homogeneous_page(
   const eamgc_homogeneous_page_t page){
  const integer_t objects_no = (integer_t)page->objects_per_page;
  integer_t i;
  for(i = 0; i < objects_no; i++)
    if(is_index_marked_in_homogeneous_page(i, page)){
#ifdef ENABLE_ASSERTIONS
      fprintf(stderr, "@@ WARNING: ");
      eamgc_dump_homogeneous_object(index_to_object_in_homogeneous_page(i, page),
                              stderr);
      fprintf(stderr, " is marked when it shouldn't be.\n");
#endif
      return (integer_t)0;
    } // if
  return (integer_t)1;
}

/* 'Reset' a possibly full homogeneous page, so that at the next allocation
   we start back considering the first object. All alive objects in the page
   must be already marked when this function is called. O(1). */
void eamgc_reset_homogeneous_page(const eamgc_homogeneous_page_t page){
  /* At the next allocation we're going to start back from the first
     object: */
  page->allocation_index = 0;
  
  //if(page->alive_objects_no == 0u)
  //  fprintf(stderr, "This page could be deleted!\n");
  
  /* We state there are no alive objects now. We're going to increment this
     while marking and finding alive objects belonging to this page: */
  page->alive_objects_no = 0u;
  
  /* Of course we don't touch the GC bits. They will be needed at
     allocation time for knowing what *not* to allocate! */
}

/* Call the appropriate trace function for object's children; object must 
   be an homogeneous object belonging to the given page (which is not checked
   for): */
void eamgc_call_trace_function_for_homogeneous_object_children(
     const word_t object,
     const eamgc_homogeneous_page_t page){
#ifdef ENABLE_ASSERTIONS
  /* Check that the object really is what it's supposed to be: */
  assert(page == eamgc_object_to_actual_homogeneous_page(object));
#endif
  
  /* Call the function: */
  page->trace_function(object);
}

void eamgc_call_for_every_possibly_nonfull_homogeneous_page(
    void (*function)(eamgc_homogeneous_page_t p)){
  /* For every pool... */
  const struct eam_doubly_linked_list_element *p; // iterator on pools
  for(p = eam_first_element(eamgc_get_all_homogeneous_pools()); p != NULL; p = eam_next(p)){
    const struct eamgc_homogeneous_pool *pool = (struct eamgc_homogeneous_pool*)eam_datum(p);
    const struct eam_doubly_linked_list_element *p; // iterator on pages
    
    /* Work on each possibly nonfull page of the pool: */
    for(p = eam_first_element(pool->possibly_nonfull_pages); p != NULL; p = eam_next(p))
      function((eamgc_homogeneous_page_t)eam_datum(p));

    /* Also work on the current page, if any: */
    if(pool->current_page != NULL)
      function(pool->current_page);
  } // outer for
}
void eamgc_call_for_every_full_homogeneous_page(
    void (*function)(eamgc_homogeneous_page_t p)){
  /* For every pool... */
  const struct eam_doubly_linked_list_element *p; // iterator on pools
  for(p = eam_first_element(eamgc_get_all_homogeneous_pools()); p != NULL; p = eam_next(p)){
    const struct eamgc_homogeneous_pool *pool = (struct eamgc_homogeneous_pool*)eam_datum(p);
    const struct eam_doubly_linked_list_element *p; // iterator on pages
    
    /* Work on each full page of the pool: */
    for(p = eam_first_element(pool->full_pages); p != NULL; p = eam_next(p))
      function((eamgc_homogeneous_page_t)eam_datum(p));
  } // outer for
}
void eamgc_call_for_every_current_homogeneous_page(
    void (*function)(eamgc_homogeneous_page_t p)){
  /* For every pool... */
  const struct eam_doubly_linked_list_element *p; // iterator on pools
  for(p = eam_first_element(eamgc_get_all_homogeneous_pools()); p != NULL; p = eam_next(p)){
    const struct eamgc_homogeneous_pool *pool = (struct eamgc_homogeneous_pool*)eam_datum(p);
    
    /* Work on the current page, if any: */
    if(pool->current_page != NULL)
      function(pool->current_page);
  } // outer for
}

/* For every existing homogeneous page (from any pool) call the given function
   passing it the page as its only parameter. An easy way to iterate (higher
   order programming in C... I must be *way* too tired). */
void eamgc_call_for_every_homogeneous_page(
         void (*function)(eamgc_homogeneous_page_t p)){
  /* For every pool... */
  const struct eam_doubly_linked_list_element *p; // iterator on pools
  for(p = eam_first_element(eamgc_get_all_homogeneous_pools()); p != NULL; p = eam_next(p)){
    /* ...work on all the pages of the pool: */
    struct eamgc_homogeneous_pool *pool = (struct eamgc_homogeneous_pool*)eam_datum(p);
    const struct eam_doubly_linked_list_element *p; // iterator on pages
    
    /* Work on the current page, if any: */
    if(pool->current_page != NULL)
      function(pool->current_page);
    
    /* Work on each possibly nonfull page: */
    for(p = eam_first_element(pool->possibly_nonfull_pages); p != NULL; p = eam_next(p))
      function((eamgc_homogeneous_page_t)eam_datum(p));
    
    /* Work on each possibly full page: */
    for(p = eam_first_element(pool->full_pages); p != NULL; p = eam_next(p))
      function((eamgc_homogeneous_page_t)eam_datum(p));
  } // outer for
}

/* Implementation of a default trace function (which traces nothing): */
static void homogeneous_trace_nothing_function_(const word_t object){
#ifdef VERBOSE_DEBUG
  fprintf(stderr, "trace_nothing_function_(): NOT tracing %p's children\n", object);
#endif // #ifdef VERBOSE_DEBUG
}
const eamgc_homogeneous_trace_function_t eamgc_homogeneous_trace_nothing_function =
  homogeneous_trace_nothing_function_;

/* Implementation of a default trace function (which, possibly as an 
   approximation, traces every word): */
static void homogeneous_trace_everything_function_(const word_t object){
#ifdef VERBOSE_DEBUG
  //  fprintf(stderr, "tracing %p's children\n", object);
#endif
  /* Infer the object size from its page header: */
  const eamgc_homogeneous_page_t homogeneous_page =
    eamgc_object_to_hypothetical_homogeneous_page(object);
#ifdef ENABLE_ASSERTIONS
  assert(homogeneous_page == eamgc_object_to_actual_homogeneous_page(object));
#endif
  const integer_t object_size = (integer_t)
    eamgc_homogeneous_page_to_object_size(homogeneous_page);
  /* Trace each word of the object assuming it's a pointer: */
  integer_t i;
  for(i = 0; i < object_size; i++)
    eamgc_trace_nth_word(object, i);
}
const eamgc_homogeneous_trace_function_t eamgc_homogeneous_trace_everything_function =
  homogeneous_trace_everything_function_;

struct eamgc_homogeneous_pool* eamgc_create_homogeneous_pool(
    const unsigned_integer_t object_size,
    const unsigned_integer_t tag,
    const word_t data,
    const eamgc_homogeneous_trace_function_t trace_function){
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "[Created an homogeneous pool for objects of size %i (%i objects per page)]\n",
          (int)object_size,
          (int)eamgc_homogeneous_object_size_to_object_no(object_size));
#endif // #ifdef ENABLE_ASSERTIONS
  /* Allocate the pool to be returned: */
  struct eamgc_homogeneous_pool *pool = (struct eamgc_homogeneous_pool*)
    eam_xmalloc(sizeof(struct eamgc_homogeneous_pool));
  
  /* Fill all fields set by the user: */
  pool->object_size = object_size;
  pool->tag = tag;
  pool->data = data;
  pool->trace_function = trace_function;
#ifdef ENABLE_ASSERTIONS
  assert(pool->trace_function != NULL);
#endif
  
  /* Compute once and for all how many object each page will be able to 
     hold: */
  pool->objects_per_page =
    eamgc_homogeneous_object_size_to_object_no(object_size);
  
  /* Make the lists for pages: */
  pool->possibly_nonfull_pages = eam_create_doubly_linked_list();
  pool->full_pages = eam_create_doubly_linked_list();
  
  /* Make a single possibly non-full page (in fact empty, in this case) for
     the new pool: */
  pool->current_page = eamgc_create_homogeneous_page(pool);
  
  /* Register the new pool, so that it behaves correctly at termination 
     time: */
  register_homogeneous_pool(pool);
  
  /* Return a pointer to the structure we built: */
  return pool;
}

/* Allocate an object from a nonfull page in the pool, if one exists; otherwise
   return NULL and set pool->current_page to NULL (we have no nonfull pages in
   that case). */
inline static word_t allocate_from_pool_without_creating_new_pages(
    struct eamgc_homogeneous_pool *pool)
  __attribute__ (( malloc ));
inline static word_t allocate_from_pool_without_creating_new_pages(
    struct eamgc_homogeneous_pool *pool){
  /* The object to return: */
  word_t object;
  
  /* Try allocating from the current page; this will succeed most of the
     times, because pages are (intentionally) quite large, and (hopefully)
     mostly empty: if allocation succeeds then we're done. Here it's
     important that GCC doesn't guess branch prediction wrong. It's 
     probably worth the trouble to optimize at least this source file
     with profile-driven feedback: */
  object = eamgc_allocate_from_homogeneous_page(pool->current_page);
  if(object != NULL)
    return object;

#ifdef VERBOSE_DEBUG
  fprintf(stderr, "( - a homogeneous page is now full - )\n");
#endif // #ifdef VERBOSE_DEBUG

  /* We weren't lucky, the current page is now full: move that into the list
     of full pages: */
  eam_prepend_to_doubly_linked_list(pool->current_page, pool->full_pages);
  
  //////////////////////////////////////////
  /* If we find some empty pages we can destroy (before a non-empty page we can
     use or before the last one), do it now, while we're looking at homogeneous
     pages. But always keep at least one page, even if empty; otherwise we
     would need to create a new one right after having deleted the last one: */
  while(eamgc_should_we_shrink_the_homogeneous_heap() &&
        (eam_doubly_linked_list_length(pool->possibly_nonfull_pages) > 1)){
    struct eamgc_homogeneous_page_header *page =
      (struct eamgc_homogeneous_page_header*)
      eam_datum(eam_first_element(pool->possibly_nonfull_pages));
    if(page->alive_objects_no == 0){// found an empty page: destroy it and go on
#ifdef VERBOSE_DEBUG
      fprintf(stderr, "---- Destroying an homogeneous page at %p!\n", page);
#endif // #ifdef VERBOSE_DEBUG
      eam_detach_first_from_doubly_linked_list(pool->possibly_nonfull_pages);
      eamgc_destroy_homogeneous_page(page);
    } // if
    else
      break; // we found the first non-empty page.
  } // while
  //////////////////////////////////////////

  /* If we have some possibly nonfull pages left then make
     one of them the new current page, and try again with that. Note that
     the recursive call is in a tail context: */
  if(! eam_is_doubly_linked_list_empty(pool->possibly_nonfull_pages)){
    pool->current_page =
      eam_detach_first_from_doubly_linked_list(pool->possibly_nonfull_pages);
#ifdef VERBOSE_DEBUG
    fprintf(stderr, "( - using the next possibly nonfull page - )\n");
#endif // #ifdef VERBOSE_DEBUG
    eamgc_reset_homogeneous_page(pool->current_page);
    return allocate_from_pool_without_creating_new_pages(pool);
  } // if
  
  /* The "last" current page is now in the list of full pages, where it 
     belongs: */
  pool->current_page = NULL;

  /* This pool has no nonempty pages. Fail: */
#ifdef VERBOSE_DEBUG
  fprintf(stderr, "( - no nonfull pages are left! - )\n");
#endif // #ifdef VERBOSE_DEBUG
  return NULL;
}

word_t eamgc_allocate_from_homogeneous_pool(struct eamgc_homogeneous_pool *pool){
  /* Try allocating without adding any new pages to the pool. In most cases
     this will succeed. Correct branch prediction in this test is important
     for performance, hence we should use profile-driven feedback to compile
     this file: */
  word_t object = allocate_from_pool_without_creating_new_pages(pool);
  if(object != NULL)
    return object;

  /* We weren't lucky, this pool is totally full. We have two opportunities
     left: either GC now and try again, or create a new page: */
#ifdef ENABLE_ASSERTIONS
    assert(pool->current_page == NULL);
#endif
  if(eamgc_is_garbage_collector_enabled() &&
     ! eamgc_should_we_enlarge_the_homogeneous_heap()){
    /* Collect and try again. Note that in some pathological cases where not
       even a single object in this pool is dead allocation may fail again, 
       even now. In such cases all we can do is just falling back to the
       other case, and create a new page: */
    eamgc_garbage_collect();
#ifdef ENABLE_ASSERTIONS
    assert(pool->current_page != NULL);
#endif
    object = allocate_from_pool_without_creating_new_pages(pool);
    if(object != NULL)
      return object;
    
    /* No space is available even right after a collection. The heap seems
       to be growing. We have nothing to do, let it grow. */
#ifdef ENABLE_ASSERTIONS
    fprintf(stderr, "+++ No space right after a GC. +++\n");
#endif
  } // if
  
#ifdef ENABLE_ASSERTIONS
  assert(pool->current_page == NULL);
#endif
  /* We have to create a new page, and try again. Allocation from the new
     page will fail only in *very* pathological cases, where fake pointers
     prevent us from finding even a single alive object. In those very rare
     cases we end up back here from the recursive call, creating yet another
     page (and possibly collecting). This can, in theory, occour more than
     once. Note that at this point pool->current_page can either be NULL or
     non-NULL, but in any case it is not what we want: it's either nothing
     or a now full page, already in the right list: we can safely overwrite
     the pointer. */
  #ifdef VERBOSE_DEBUG
  fprintf(stderr, "( - creating a new homogeneous page - )\n");
  #endif // #ifdef VERBOSE_DEBUG
  pool->current_page = eamgc_create_homogeneous_page(pool);
  return eamgc_allocate_from_homogeneous_pool(pool);
}

/* Go back considering all allocated space as potentially containing dead
   objects. This must be called only by the collector right after marking,
   so that no information about which objects are still alive is lost.
   O(1). */
void eamgc_reset_homogeneous_pool(struct eamgc_homogeneous_pool *pool){
  /* The pages which were full before now can have free space: */
  pool->possibly_nonfull_pages = 
    eam_concatenate_doubly_linked_lists(pool->possibly_nonfull_pages,
                                    pool->full_pages);
  
  /* There are no full pages anymore: */
  pool->full_pages = eam_create_doubly_linked_list();
  
  /* If there is a current page (this may not be the case, because we're
     resetting the pool as part of a collection) then reset it, and keep
     that as the current page; otherwise pick a new current page from the 
     just-updated list of possibly non-full pages, (it *must* exist
     because each pool must always have at least one page), and reset
     that: */
  if(pool->current_page == NULL){
#ifdef ENABLE_ASSERTIONS
    assert(! eam_is_doubly_linked_list_empty(pool->possibly_nonfull_pages));
#endif
    pool->current_page = 
      eam_detach_first_from_doubly_linked_list(pool->possibly_nonfull_pages);
  } // if
  eamgc_reset_homogeneous_page(pool->current_page);
  
  /* Ok, we have a current page, a (now possibly empty) list of possibly
     non-full pages, and an empty list of full pages. This pool is ready
     for allocating again. */
}

/* Overwrite all unmarked objects in the page with the given pattern. Useful 
   for debugging. Pages are not moved from full to possibly non-full.
   Of course this should be called only after a complete marking.  */
void eamgc_overwrite_dead_objects_in_homogeneous_pool(
   struct eamgc_homogeneous_pool *pool,
   const word_t pattern){
  struct eam_doubly_linked_list_element *p;
  for(p = eam_first_element(pool->possibly_nonfull_pages); p != NULL; p = eam_next(p))
    eamgc_overwrite_dead_objects_in_homogeneous_page((eamgc_homogeneous_page_t)eam_datum(p));
  for(p = eam_first_element(pool->full_pages); p != NULL; p = eam_next(p))
    eamgc_overwrite_dead_objects_in_homogeneous_page((eamgc_homogeneous_page_t)eam_datum(p));
  assert(pool->current_page != NULL);
  eamgc_overwrite_dead_objects_in_homogeneous_page(pool->current_page);
}

/* Unmark all the objects in all the homogeneous pages of the pool. */
void eamgc_unmark_all_objects_in_homogeneous_pool(struct eamgc_homogeneous_pool *pool){
  struct eam_doubly_linked_list_element *p;
  /* Unmark all objects in the possibly nonfull pages: */
  for(p = eam_first_element(pool->possibly_nonfull_pages); p != NULL; p = eam_next(p))
    eamgc_unmark_all_objects_in_homogeneous_page((eamgc_homogeneous_page_t)eam_datum(p));

  /* Unmark all objects in the full pages: */
  for(p = eam_first_element(pool->full_pages); p != NULL; p = eam_next(p))
    eamgc_unmark_all_objects_in_homogeneous_page((eamgc_homogeneous_page_t)eam_datum(p));

  /* Unmark all objects in the current page, if any: */
  if(pool->current_page != NULL)
    eamgc_unmark_all_objects_in_homogeneous_page(pool->current_page);
}

integer_t eamgc_are_all_objects_unmarked_in_homogeneous_pool(
    const struct eamgc_homogeneous_pool *pool){
  /* If any single page has a marked bit then we return zero, else we
     can be sure no page has marks, and return non-zero: */
  struct eam_doubly_linked_list_element *p;
  
  /* Check possibly nonfull pages: */
  for(p = eam_first_element(pool->possibly_nonfull_pages); p != NULL; p = eam_next(p))
    if(! eamgc_are_all_objects_unmarked_in_homogeneous_page(eam_datum(p)))
      return 0; 
  
  /* Check full pages: */
  for(p = eam_first_element(pool->full_pages); p != NULL; p = eam_next(p))
    if(! eamgc_are_all_objects_unmarked_in_homogeneous_page(eam_datum(p)))
      return 0; 
  
  /* Check the current page, if any: */
  if(pool->current_page != NULL)
    if(! eamgc_are_all_objects_unmarked_in_homogeneous_page(pool->current_page))
      return 0;
  
  /* Ok, every page we checked is fine: */
  return 1;
}

/* For every existing pool call the given function passing it the pool
   as its only parameter. An easy way to iterate (higher order programming
   in C... I must be *way* too tired). */
void eamgc_call_for_every_homogeneous_pool(
    void (*function)(struct eamgc_homogeneous_pool *pool)){
  const struct eam_doubly_linked_list_element *p; // iterator on pools
  for(p = eam_first_element(eamgc_get_all_homogeneous_pools()); p != NULL; p = eam_next(p))
    function((struct eamgc_homogeneous_pool*)eam_datum(p));
}
