/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with epsilon; see the file COPYING.  If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA. */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "large.h"
#include "trace.h"
#include "compile_time_parameters.h"
#include "statistics.h"
#include "heuristics.h"
#include "../../common/malloc.h"
#include "../../common/list.h"
#include "../../common/set.h"

/* Don't trace any of the large object children: */
static void eamgc_large_trace_nothing_function_(const word_t large_object,
                                             const unsigned_integer_t size){
  /* Do nothing. */
}
const eamgc_large_trace_function_t eamgc_large_trace_nothing_function =
  eamgc_large_trace_nothing_function_;

/* Trace every word of the given large object as a children of its: */
static void eamgc_large_trace_everything_function_(const word_t large_object,
                                                const unsigned_integer_t size){
  /* Trace each word of the object assuming it's a pointer: */
  unsigned_integer_t i;
  for(i = 0u; i < size; i++)
    eamgc_trace_nth_word(large_object, i);
}
const eamgc_large_trace_function_t eamgc_large_trace_everything_function =
  eamgc_large_trace_everything_function_;

/* Default "finalization" for a large object, doing nothing: */
static void eamgc_default_large_finalize_function_(const word_t large_object,
                                                const unsigned_integer_t size){
  /* Do nothing. */
}
const eamgc_large_finalize_function_t eamgc_default_large_finalize_function =
  eamgc_default_large_finalize_function_;

/* The header describing a single large object. Note that the header size
   is multiple of a word size, hence all large objects are (at least)
   word-aligned: */
struct eamgc_large_object_header{
  /* How much this object is large, in words: */
  unsigned_integer_t object_size;
  
  /* The pool, containing other information which is common to all objects
     from the same pool: */
  struct eamgc_large_pool* pool;

  /* The single mark 'bit' for the object. */
  unsigned_integer_t mark_bit;
}; // struct

/* A large object is always allocated right after its header; it's easy to
   find the header of a given object, and vice-versa, by adding or
   substracting a fixed offset: */
typedef struct eamgc_large_object_header* large_object_with_header_t;

/* The structure holding together "similar" large objects: */
struct eamgc_large_pool{
  /* List of all large objects headers (so they are pointers to headers,
     and *not* to actual objects): */
  struct eam_list *objects_with_headers;
  
  /* Total size, in words, of all the currently existing large objects
     of this pool: */
  unsigned_integer_t total_objects_size;
  
  /* Number of existing large objects in this pool: */
  unsigned_integer_t objects_no;
  
  /* The tag of all the objects of this pool: */
  unsigned_integer_t tag;
  
  /* The data associated to all objects from this pool (in analogy to 
     what happens with homogeneous pools: */
  word_t data;
  
  /* The procedure used to mark the objects referred by each object of this
     pool: this must not be NULL. */
  eamgc_large_trace_function_t trace_function;
  
  /* The procedure used to finalize an object of this pool, or NULL if objects
     from this pool do not require finalization: */
  eamgc_large_finalize_function_t finalize_function;
}; // struct

/* The associative data structure set used to quickly find whether a given
   large object exists: */
static set_t set_of_large_objects;

/* Global statistics about all the large objects from all pools: */
static unsigned_integer_t total_large_objects_no;
static unsigned_integer_t total_large_objects_size;

/* Given a pointer to a large object return a pointer to its header. This
   should not cost more than one assembler instruction: */
#define LARGE_OBJECT_TO_HEADER(o) \
  ((struct eamgc_large_object_header*) \
   (((char*)(o)) - sizeof(struct eamgc_large_object_header)))

/* Given a pointer to a large object header return a pointer to the large
   object itself. This should not cost more than one assembler instruction: */
#define HEADER_AND_LARGE_OBJECT_TO_LARGE_OBJECT(oh) \
  ((word_t) \
   (((char*)(oh)) + sizeof(struct eamgc_large_object_header)))

/* The list of all existing large pools: */
static struct eam_list *eamgc_large_pools;

/* Initialize the data structures needed to handle large objects: */
void eamgc_initialize_large_objects_structures(void){
  /* Initialize the list of large pools: */
  eamgc_large_pools = eam_nil;

  /* Create the set of large objects: */
  set_of_large_objects = eam_create_set(SIZE_OF_SET_OF_LARGE_OBJECTS);
  
  /* Initialize global statistics data: */
  total_large_objects_no = 0u;
  total_large_objects_size = 0u;
}

/* *Not* for the user! This assumes some data structures have been already
   destroyed. */
static void destroy_large_pool(struct eamgc_large_pool *pool){
  /* Free what remains of the list: */
  eam_destroy_list_structure(pool->objects_with_headers);

  /* Free the buffer: */
  free(pool);
}

/* Finalize the data structures needed to handle large objects: */
void eamgc_finalize_large_objects_structures(void){
  /* Finalize and destroy allocated objects: */
  eamgc_finalize_and_destroy_large_objects();
  
  /* Destroy pools: */
  eamgc_call_for_every_large_pool(destroy_large_pool);
  
  /* Destroy what remains of the list of large pools: */
  eam_destroy_list_structure(eamgc_large_pools);
  
  /* Destroy the set of large objects: */
  eam_destroy_set(set_of_large_objects);
}

/* Return a new large pool with the given features: */
struct eamgc_large_pool* eamgc_create_large_pool(
    const unsigned_integer_t tag,
    const word_t pool_data,
    const eamgc_large_trace_function_t trace_function,
    const eamgc_large_finalize_function_t finalize_function){
#ifdef ENABLE_ASSERTIONS
  /* Check that the trace functions is not NULL. This is an easy error to make... */
  assert(trace_function != NULL);
#endif // #ifdef ENABLE_ASSERTIONS
  
#ifdef ENABLE_VERBOSE_DEBUG
  fprintf(stderr, "[Created a large pool]\n");
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* Allocate a pool: */
  struct eamgc_large_pool* pool =
    eam_xmalloc(sizeof(struct eamgc_large_pool));
  
  /* Fill its fields (there are no objects yet, of course): */
  pool->objects_with_headers = eam_nil;
  pool->total_objects_size = 0u;
  pool->objects_no = 0u;
  pool->tag = tag;
  pool->data = pool_data;
  pool->trace_function = trace_function;
  pool->finalize_function = finalize_function;
  
  /* Add the new pool to the global list of pools: */
  eamgc_large_pools = eam_cons(pool, eamgc_large_pools);
  
  /* Return the pool we made: */
  return pool;
}

/* Create a new large object from the given large objects pool. *Never*
   trigger a GC. Of course this is only internally used, from the function
   exposed to the user: */
static word_t eamgc_always_allocate_from_large_pool(struct eamgc_large_pool* pool,
                                                 const unsigned_integer_t object_size)
  __attribute__(( malloc ));
static word_t eamgc_always_allocate_from_large_pool(struct eamgc_large_pool* pool,
                                                 const unsigned_integer_t object_size){
  /* Allocate a buffer large enough to hold the header *and* the object. Note
     that we don't need to care about page-alignment here: */
  large_object_with_header_t header_and_object = (large_object_with_header_t)
    eam_xmalloc(sizeof(struct eamgc_large_object_header) + 
            object_size * sizeof(word_t));
  word_t object = HEADER_AND_LARGE_OBJECT_TO_LARGE_OBJECT(header_and_object);
  
  /* Fill the header: */
  header_and_object->object_size = object_size;
  header_and_object->pool = pool;
  header_and_object->mark_bit = 0u;
  
  /* Add the object header to the list of objects of its pool, and the object
     to the global set of large objects: */
  pool->objects_with_headers = eam_cons(header_and_object, 
                                        pool->objects_with_headers);
  eam_insert_pointer_into_set(object, set_of_large_objects);
  
  /* We added an object which is object_size words wide: */
  pool->objects_no++;
  pool->total_objects_size += object_size;
  total_large_objects_no++;
  total_large_objects_size += object_size;
  
  /* Return the object (*not* the whole buffer!): */
  return object;
}

/* Perform a collection if needed, then allocate a new object from the
   given pool: */
word_t eamgc_allocate_from_large_pool(struct eamgc_large_pool* pool,
                                   const unsigned_integer_t object_size){
  /* Should we collect? If so do it now: */
  if(eamgc_is_garbage_collector_enabled() &&
     ! eamgc_should_we_enlarge_the_large_heap()){
    eamgc_garbage_collect();
    /* Note that we don't know how much space was freed by the collection;
       in any case there's nothing we can do about it (we *need* to allocate,
       and we've just collected) so we don't even bother to check. */
  } // if
  
  /* Allocate the requested object: */
  return eamgc_always_allocate_from_large_pool(pool, object_size);
}

/* Destroy the given large object, removing the reference to it from the set
   of large objects (the list from the pool is *not* touched by this function).
   This doesn't check that large_object actually is a large object: */
static void eamgc_destroy_large_object(const word_t large_object){
#ifdef ENABLE_ASSERTIONS
  /* Check that lo actually is a large object: */
  assert(eamgc_is_an_actual_large_object(large_object));
#endif // #ifdef ENABLE_ASSERTIONS
  
  struct eamgc_large_object_header* header =
    LARGE_OBJECT_TO_HEADER(large_object);
  const unsigned_integer_t object_size = header->object_size;
  struct eamgc_large_pool* pool = header->pool;
  
  /* Destroy the object and its header, which were allocated as a single
     buffer: */
  free(header);
  
  /* We destroyed an object which was object_size words wide: */
  pool->objects_no--;
  pool->total_objects_size -= object_size;
  total_large_objects_no--;
  total_large_objects_size -= object_size;
  
  /* Remove the reference from the set: */
  eam_remove_pointer_from_set(large_object, set_of_large_objects);
}

/* Return a nonzero value iff the given pointer refers an alive large
   object: */
unsigned_integer_t eamgc_is_an_actual_large_object(const word_t large_object){
  return (unsigned_integer_t)
    eam_does_pointer_belong_to_set(large_object, set_of_large_objects);
}

/* Return the size (in words) of the given large object, which must exist: */
unsigned_integer_t eamgc_large_object_to_size(const word_t lo){
#ifdef ENABLE_ASSERTIONS
  /* Check that lo actually is a large object: */
  assert(eamgc_is_an_actual_large_object(lo));
#endif // #ifdef ENABLE_ASSERTIONS

  /* Return what we need (from the header): */
  return LARGE_OBJECT_TO_HEADER(lo)->object_size;
}

/* Return the tag of the given large object, which must exist: */
unsigned_integer_t eamgc_large_object_to_tag(const word_t lo){
#ifdef ENABLE_ASSERTIONS
  /* Check that lo actually is a large object: */
  assert(eamgc_is_an_actual_large_object(lo));
#endif // #ifdef ENABLE_ASSERTIONS

  /* Return what we need (from the pool): */
  return LARGE_OBJECT_TO_HEADER(lo)->pool->tag;
}

/* Return the data of the given large object, which must exist: */
word_t eamgc_large_object_to_data(const word_t lo){
#ifdef ENABLE_ASSERTIONS
  /* Check that lo actually is a large object: */
  assert(eamgc_is_an_actual_large_object(lo));
#endif // #ifdef ENABLE_ASSERTIONS

  /* Return what we need (from the pool): */
  return LARGE_OBJECT_TO_HEADER(lo)->pool->data;
}

/* Return the data of the given large pool: */
inline word_t eamgc_large_pool_to_data(const struct eamgc_large_pool *pool){
  return pool->data;
}

void eamgc_call_trace_function_for_large_object_children(const word_t large_object){
#ifdef ENABLE_ASSERTIONS
  /* Check that large_object actually is a large object: */
  assert(eamgc_is_an_actual_large_object(large_object));
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* Find a pointer to the function and invoke it: */
  const struct eamgc_large_object_header* header =
    LARGE_OBJECT_TO_HEADER(large_object);
  header->pool->trace_function(large_object, header->object_size);
}

/* Set the mark bit of the given large object, which must exist (this 
   is not checked for). The previous value of the mark bit is returned: */
static unsigned_integer_t eamgc_set_large_object_mark_bit(const word_t large_object,
                                                       const unsigned_integer_t value)
  __attribute__(( always_inline ));
static unsigned_integer_t eamgc_set_large_object_mark_bit(const word_t large_object,
                                                          const unsigned_integer_t value){
#ifdef ENABLE_ASSERTIONS
  /* Check that large_object actually is a large object: */
  assert(eamgc_is_an_actual_large_object(large_object));
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* Obtain the header from the object: */
  struct eamgc_large_object_header* header =
    LARGE_OBJECT_TO_HEADER(large_object);
  
  /* Set the new value, without forgetting the previous value: */
  const unsigned_integer_t previous_value = header->mark_bit;
  header->mark_bit = value;
  
  /* If we marked a previously unmarked object then we found an alive
     object: */
  if((previous_value == 0) && (value != 0))
    eamgc_increment_alive_large_heap_size(header->object_size);
  
  /* Return the previous value that we saved before: */
  return previous_value;
}

/* Mark the given large object, which must exist (this is not checked for). 
   The previous value of the mark bit is returned: */
integer_t eamgc_mark_large_object(const word_t large_object){
  return eamgc_set_large_object_mark_bit(large_object, 1);
}

/* Clear the mark bit of the given large object, which must exist (this is
   not checked for): */
void eamgc_unmark_large_object(const word_t large_object){
  /* Any decent optimizer will avoid to generate code computing the ignored
     return value in this case; and GCC is quite a lot more than decent: */
  eamgc_set_large_object_mark_bit(large_object, 0);
}

/* Return the current value of the mark bit of the given object, which must
   exist (this is not checked for): */
integer_t eamgc_is_large_object_marked(const word_t large_object){
#ifdef ENABLE_ASSERTIONS
  /* Check that large_object actually is a large object: */
  assert(eamgc_is_an_actual_large_object(large_object));
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* Return the mark bit value, taken from the header: */
  return LARGE_OBJECT_TO_HEADER(large_object)->mark_bit;
}

/* Return the total memory space (in words) used by large objects payloads.
   Headers, GC bits and other data structures are not counted: */
inline unsigned_integer_t eamgc_get_large_heap_size(void){
  return total_large_objects_size;
}

/* For every existing pool call the given function passing it the pool
   as its only parameter. An easy way to iterate. */
void eamgc_call_for_every_large_pool(void (*function)(struct eamgc_large_pool *pool)){
  const struct eam_list *p; // iterator on pools
  for(p = eamgc_large_pools; p != NULL; p = eam_cdr(p))
    function((struct eamgc_large_pool*)eam_car(p));
}

/* Finalize and destroy all unmarked large objects, and unmark the marked
   ones: */
static void eamgc_sweep_large_pool(struct eamgc_large_pool *pool){
#ifdef ENABLE_ASSERTIONS
  /* This should always be called with the GC (temporarily) disabled: */
  assert(! eamgc_is_garbage_collector_enabled());
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* Get the finalizer, once and for all: */
  const eamgc_large_finalize_function_t finalizer = pool->finalize_function;

  /* Scan the list of large objects: destroy dead objects and make a new
     list with the alive ones (while unmarking them). It's important that
     before looping we clear the list of large objects, because finalizations
     may trigger allocations *also in this same pool*: hence it's possible
     that some new objects are prepended to the new list while we scroll the
     old one: */
  struct eam_list *old_list = pool->objects_with_headers;
  pool->objects_with_headers = eam_nil;
  struct eam_list *p;
  for(p = old_list ; ! eam_null(p); p = eam_cdr(p)){
    struct eamgc_large_object_header *header = eam_car(p);
    const word_t object = HEADER_AND_LARGE_OBJECT_TO_LARGE_OBJECT(header);
    /* Is the object currently marked? */
    if(eamgc_is_large_object_marked(object)){
      /* The object is alive; unmark it, and add it to the new list: */
      eamgc_unmark_large_object(object);
      pool->objects_with_headers = eam_cons(header, pool->objects_with_headers);
    }
    else{
      /* The object is dead. Finalize it if needed, deallocate it and remove it
         from the set. Don't add it to the new list, of course: */
      if(finalizer != NULL) // GCC moves this check out of the loop, of course
        finalizer(object, header->object_size);
      eamgc_destroy_large_object(object);
    } // else
  } // for

  /* We scanned the whole list: destroy all its conses; the still alive elements
     are shared and still needed, so we don't touch them: */
  eam_destroy_list_structure(old_list);
  
  /* Ok, the new list is now in place, and may already contain some new
     objects created during finalization. */
}

/* Finalize and destroy all unmarked large objects, and unmark the marked
   ones: */
void eamgc_sweep_large_objects(void){
  /* Temporarily disable GC: finalization may trigger allocation, which may
     trigger another GC; this would be complex to manage because of the way
     we work with lists. It's better to tolerate some heap growth: */
  const integer_t was_gc_enabled = eamgc_is_garbage_collector_enabled();
  eamgc_disable_garbage_collector();
  
  /* Sweep all pools: */
  eamgc_call_for_every_large_pool(eamgc_sweep_large_pool);
  
  /* Now we can re-enable GC, if it was on before: */
  if(was_gc_enabled)
    eamgc_enable_garbage_collector();
}

/* Finalize all the large objects of the given pool, without destroying them.
   Only useful at exit time: */
static void eamgc_finalize_objects_from_large_pool(struct eamgc_large_pool *pool){
  /* Get the finalizer, once and for all. Do nothing if we don't need
     finalization for this pool: */
  const eamgc_large_finalize_function_t finalizer = pool->finalize_function;
  if(finalizer == NULL)
    return;
  
  /* For each large object... */
  struct eam_list *p;
  for(p = pool->objects_with_headers; ! eam_null(p); p = eam_cdr(p)){
    const struct eamgc_large_object_header *header =
      (const struct eamgc_large_object_header*)eam_car(p);
    word_t object = HEADER_AND_LARGE_OBJECT_TO_LARGE_OBJECT(header);
    
    /* ...finalize it (finalizer is not NULL if we are here): */
    finalizer(object, header->object_size);
  } // for
}

/* Finalize all existing large objects, without destroying them. Only useful
   at exit time: */
void eamgc_finalize_large_objects(void){
  /* Temporarily disable GC: finalization may trigger allocation, which may
     trigger another GC; this would be complex to manage because of how we
     work with lists. It's better to tolerate some heap growth: */
  const integer_t was_gc_enabled = eamgc_is_garbage_collector_enabled();
  eamgc_disable_garbage_collector();
  
  /* Finalize the objects of all pools: */
  eamgc_call_for_every_large_pool(eamgc_finalize_objects_from_large_pool);
  
  /* Now we can re-enable GC, if it was on before: */
  if(was_gc_enabled)
    eamgc_enable_garbage_collector();
}

/* Finalize and destroy all existing large objects: */
static void eamgc_finalize_and_destroy_objects_from_large_pool(
    struct eamgc_large_pool *pool){
  /* Get the finalizer, once and for all: */
  const eamgc_large_finalize_function_t finalizer = 
    pool->finalize_function;
  
  /* For each object... */
  const struct eam_list *p;
  for(p = pool->objects_with_headers; ! eam_null(p); p = eam_cdr(p)){
    struct eamgc_large_object_header *header =
      (struct eamgc_large_object_header*)eam_car(p);
    word_t object = HEADER_AND_LARGE_OBJECT_TO_LARGE_OBJECT(header);

    /* ...finalize it and destroy it: */
    if(finalizer != NULL) // GCC optimizer moves this test out of the loop
      finalizer(object, header->object_size);
    eamgc_destroy_large_object(object);
  } // for
}

/* Finalize and destroy all existing large objects: */
void eamgc_finalize_and_destroy_large_objects(void){
  /* Temporarily disable GC: finalization may trigger allocation, which may
     trigger another GC; this would be complex to manage because of how we
     work with lists. It's better to tolerate some heap growth: */
  const integer_t was_gc_enabled = eamgc_is_garbage_collector_enabled();
  eamgc_disable_garbage_collector();
  
  /* Work on every pool: */
  eamgc_call_for_every_large_pool(eamgc_finalize_and_destroy_objects_from_large_pool);
  
  /* Now we can re-enable GC, if it was on before: */
  if(was_gc_enabled)
    eamgc_enable_garbage_collector();
}
