/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EAMGC_LARGE_H_
#define EAMGC_LARGE_H_

#include "../../common/types.h"

/* A possibly user-supplied function tracing a large object of some fixed
   kind. Pointer-finding may be conservative, but of course *at least* all
   real pointers must be marked. Note that the object itself should not be
   traced by this function; all its direct children should. size represents
   the object size, in words: */
typedef void (*eamgc_large_trace_function_t)(const word_t,
                                          const unsigned_integer_t size);

/* A possibly user-supplied function destroying an object of some fixed 
   kind. size represents the object size, in words:*/
typedef void (*eamgc_large_finalize_function_t)(const word_t,
                                             const unsigned_integer_t size);

/* Reasonable defaults for the function types above:
   trace_everything_function may be used for conservative pointer finding;
   Note that there is no eamgc_default_large_finalize_function; it would be
   less efficient that using NULL. */
extern const eamgc_large_trace_function_t eamgc_large_trace_nothing_function;    // trace nothing
extern const eamgc_large_trace_function_t eamgc_large_trace_everything_function; // trace every word
extern const eamgc_large_finalize_function_t eamgc_default_large_finalize_function; // do nothing

/* Initialize the data structures needed to handle large objects. The user
   never needs to directly call this: */
void eamgc_initialize_large_objects_structures(void);

/* Finalize the data structures needed to handle large objects. The user
   never needs to directly call this: */
void eamgc_finalize_large_objects_structures(void);

/* Return a new large pool with the given features. Note that finalize_function
   *can* be NULL, but trace_function can't. */
struct eamgc_large_pool* eamgc_create_large_pool(
    const unsigned_integer_t tag,
    const word_t pool_data,
    const eamgc_large_trace_function_t trace_function,       // not NULL
    const eamgc_large_finalize_function_t finalize_function) // can be NULL
  __attribute__(( malloc ));

/* Perform a collection if needed, then allocate a new object from the
   given pool: */
word_t eamgc_allocate_from_large_pool(struct eamgc_large_pool* pool,
                                   const unsigned_integer_t object_size)
  __attribute__(( malloc ));

/* Finalize and destroy all unmarked large objects, and unmark the marked
   ones: */
void eamgc_sweep_large_objects(void);

/* Finalize all existing large objects, without destroying them. Only useful
   at exit time: */
void eamgc_finalize_large_objects(void);

/* Finalize and destroy all existing large objects: */
void eamgc_finalize_and_destroy_large_objects(void);

/* Return a nonzero value iff the given pointer refers an alive large
   object: */
unsigned_integer_t eamgc_is_an_actual_large_object(const word_t lo)
  __attribute__(( pure ));

/* Return the size (in words) of the given large object, which must exist: */
unsigned_integer_t eamgc_large_object_to_size(const word_t lo)
  __attribute__(( pure ));

/* Return the tag of the given large object, which must exist: */
unsigned_integer_t eamgc_large_object_to_tag(const word_t lo)
  __attribute__(( pure ));

/* Return the data of the given large object, which must exist: */
word_t eamgc_large_object_to_data(const word_t lo)
  __attribute__(( pure ));

/* Return the data of the given large pool: */
inline word_t eamgc_large_pool_to_data(const struct eamgc_large_pool *pool)
  __attribute__(( pure ));

/* Mark the given large object, which must exist (this is not checked for). 
   The previous value of the mark bit is returned: */
integer_t eamgc_mark_large_object(const word_t large_object);

/* Clear the mark bit of the given large object, which must exist (this is
   not checked for): */
void eamgc_unmark_large_object(const word_t large_object);

/* Return the current value of the mark bit of the given object, which must
   exist (this is not checked for): */
integer_t eamgc_is_large_object_marked(const word_t large_object)
  __attribute__(( pure ));

/* Call the appropriate trace function for large_object's children;
   large_object must be an existing large object (which is not checked
   for): */
void eamgc_call_trace_function_for_large_object_children(
   const word_t large_object);

/* Return the total memory space (in words) used by large objects payloads.
   Headers, GC bits and other data structures are not counted: */
inline unsigned_integer_t eamgc_get_large_heap_size(void)
  __attribute__ (( pure ));

/* For every existing pool call the given function passing it the pool
   as its only parameter. An easy way to iterate (higher order programming
   in C... I must be *way* too tired). */
void eamgc_call_for_every_large_pool(
    void (*function)(struct eamgc_large_pool *pool));

// To do: finalization at exit time.

#endif // #ifndef EAMGC_LARGE_H_
