/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EAMGC_ROOTS_H_
#define EAMGC_ROOTS_H_

#include "gc.h"

/* Each root should be registered once, typically but not necessarily 
   at initializtion time. In particular there is no need to re-register 
   roots after every collection. There's no limit to the number of
   roots and root tracing functions.
   No facility is currently provided for unregistering roots. */

/* Initialize data structures for the roots filesystem. Not for the user: */
void eamgc_initialize_roots_structures(void);

/* Initialize data structures for the roots filesystem. Not for the user: */
void eamgc_finalize_roots_structures(void);

/* Register a user-defined root, to be traced at the beginning of every 
   collection. Note that a pointer to the buffer of roots must be 
   supplied. The scanned words will be:
   - *first_pointer_to_root
   - *((void*)first_pointer_to_root + 1 * sizeof(word_t))
   - *((void*)first_pointer_to_root + 2 * sizeof(word_t))
   - *((void*)first_pointer_to_root + 3 * sizeof(word_t))
   - ...
   roots_no is the length of the buffer, where of course each element is
   a word. */
void eamgc_register_buffer_of_roots(const word_t first_pointer_to_root,
                                 const unsigned_integer_t roots_no);

/* Register a single pointer to the heap as a root. It's equivalent to
   eamgc_register_root_buffer() with 1u as the second parameter. The heap
   pointer which is traced is (the future value of) *root_pointer. */
void eamgc_register_root(const word_t root_pointer);

/* Register a buffer already in the heap as a root. The heap pointers
   which are traced start at root (not at *root, even if that is of
   course also traced). The length is in words. */
void eamgc_register_buffer_of_heap_roots(const word_t root_beginning,
                                      const unsigned_integer_t root_length);

/* Register a single pointer already in the heap as a root. The heap pointer
   which is traced is root (not only *root, even if that is of course
   also traced). This is equivalent to eamgc_register_buffer_of_heap_roots()
   with 1u as the second parameter: */
void eamgc_register_heap_root(const word_t root);


/* A root tracing function is a typically user-supplied function used
   to trace roots. Once registered with eamgc_register_root_tracing_function()
   all root tracing functions are automatically called at the beginning
   of every collection. */
typedef void (*eamgc_root_tracing_function_t)(void);

/* Register a function so that it's called at the beginning of every
   collection to trace some user-defined roots. */
void eamgc_register_root_tracing_function(const eamgc_root_tracing_function_t function);

/* A root tracing function which the user may register if he/she desires 
   so (no root tracing function is registered by default), tracing roots
   from the machine registers and the C stack. */
extern const eamgc_root_tracing_function_t eamgc_trace_c_context_roots;


/* Trace all roots (including pool data). This is called by the collector
   to start tracing. Not for the user. */
void eamgc_trace_roots(void);

#endif // #ifndef EAMGC_ROOTS_H_
