#define _GNU_SOURCE

#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <assert.h>

typedef unsigned int unsigned_integer_t;

#ifndef WORD_BIT
#define WORD_BIT (CHAR_BIT * sizeof(unsigned_integer_t))
#endif // #ifndef WORD_BIT

void dump_word(const unsigned_integer_t w, FILE *f){
  unsigned_integer_t i;
  for(i = 0; i < WORD_BIT; i++)
    if(w & (1u << i))
      fprintf(f, "1");
    else
      fprintf(f, "0");
}

#define DO_N_TIMES_BEGIN(N) {int _q; const int _max = N; for(_q = 0; _q < _max; _q++){
#define DO_TIMES_END        }}
#define DO_MANY_TIMES_BEGIN DO_N_TIMES_BEGIN(10000000)

#define BUFFER_LENGTH_IN_WORDS 32768
struct bit_stream_outgoing_interface{
  unsigned_integer_t next_bit_to_write;
#ifdef BUFFER_LENGTH_IN_WORDS
  unsigned_integer_t buffer[BUFFER_LENGTH_IN_WORDS];
#else
  unsigned_integer_t buffer;
#endif // #ifdef BUFFER_LENGTH_IN_WORDS
  int fd;
}; // struct
typedef struct bit_stream_outgoing_interface* bit_stream_outgoing_interface_t;

#warning To do: open/close

struct bit_stream_incoming_interface{
  unsigned_integer_t next_bit_to_read;
#ifdef BUFFER_LENGTH_IN_WORDS
  unsigned_integer_t buffer[BUFFER_LENGTH_IN_WORDS];
#else
  unsigned_integer_t buffer;
#endif // #ifdef BUFFER_LENGTH_IN_WORDS
  int fd;
}; // struct
typedef struct bit_stream_incoming_interface* bit_stream_incoming_interface_t;

int flush_bit_stream_outgoing_interface(bit_stream_outgoing_interface_t i){
  /* Do nothing if there's nothing to flush: */
  if(i->next_bit_to_write == 0)
    return 0;
  
  /* Write the buffer to the output: */
  ssize_t write_return_value;
#ifdef BUFFER_LENGTH_IN_WORDS
  size_t bytes_to_write = sizeof(unsigned_integer_t) * BUFFER_LENGTH_IN_WORDS;
#else
  size_t bytes_to_write = sizeof(unsigned_integer_t);
#endif
  do{
    write_return_value =
#ifdef BUFFER_LENGTH_IN_WORDS
      TEMP_FAILURE_RETRY(write(i->fd, i->buffer, bytes_to_write));
#else
      TEMP_FAILURE_RETRY(write(i->fd, &i->buffer, bytes_to_write));
#endif
    if(write_return_value > 0)
      bytes_to_write -= write_return_value;
  }while((bytes_to_write > 0) && (write_return_value > 0));

  /* Ok, we can reuse the whole buffer now: */
  i->next_bit_to_write = 0;

  /* Clear the buffer: this enables us to only set the 1 bits, saving some
     time when writing: */
#ifdef BUFFER_LENGTH_IN_WORDS
  memset(i->buffer, 0u, BUFFER_LENGTH_IN_WORDS * sizeof(unsigned_integer_t));
#else
  i->buffer = 0u;
#endif // #ifdef BUFFER_LENGTH_IN_WORDS

  /* Did we write with success? */
  return (write_return_value > 0) ? 0 : -1;
}

int fill_bit_stream_incoming_interface(bit_stream_incoming_interface_t i){
  /* Fill the buffer from the input stream: */
  ssize_t read_return_value;
#ifdef BUFFER_LENGTH_IN_WORDS
  size_t bytes_to_read = sizeof(unsigned_integer_t) * BUFFER_LENGTH_IN_WORDS;
#else
  size_t bytes_to_read = sizeof(unsigned_integer_t);
#endif // #ifdef BUFFER_LENGTH_IN_WORDS

  do{
    read_return_value =
#ifdef BUFFER_LENGTH_IN_WORDS
      TEMP_FAILURE_RETRY(read(i->fd, i->buffer, bytes_to_read));
#else
      TEMP_FAILURE_RETRY(read(i->fd, &i->buffer, bytes_to_read));
#endif // #ifdef BUFFER_LENGTH_IN_WORDS
    if(read_return_value > 0)
      bytes_to_read -= read_return_value;
  }while((bytes_to_read > 0) && (read_return_value > 0));
  
  /* Start again reading from the beginning of the buffer: */
  i->next_bit_to_read = 0u;
  
  /* Did we read with success? */
  return (read_return_value > 0) ? 0 : -1;
}

unsigned_integer_t read_bit(bit_stream_incoming_interface_t i){
  /* If we read all the buffer then fill it with the next block: */
#ifdef BUFFER_LENGTH_IN_WORDS
  if(i->next_bit_to_read == (WORD_BIT * BUFFER_LENGTH_IN_WORDS))
#else
  if(i->next_bit_to_read == WORD_BIT)
#endif
    fill_bit_stream_incoming_interface(i);
  
  /* Ok, now we're sure there's something to read; read it: */
#ifdef BUFFER_LENGTH_IN_WORDS
  const unsigned_integer_t word_index = i->next_bit_to_read / WORD_BIT;
  const unsigned_integer_t offset = i->next_bit_to_read % WORD_BIT;
  const unsigned_integer_t bitmask = 1u << offset;
  const unsigned_integer_t bit = i->buffer[word_index] & bitmask;
#else
  const unsigned_integer_t bitmask = 1u << i->next_bit_to_read;
  const unsigned_integer_t bit = i->buffer & bitmask;
#endif  

  /* Advance the index to the next bit: */
  i->next_bit_to_read++;

  /* Return the bit we read: */
  return bit;
}

void write_bit(bit_stream_outgoing_interface_t i, const unsigned_integer_t bit){
  /* Flush if we're about to overflow: */
#ifdef BUFFER_LENGTH_IN_WORDS
  if(i->next_bit_to_write == (WORD_BIT * BUFFER_LENGTH_IN_WORDS))
#else
  if(i->next_bit_to_write == WORD_BIT)
#endif
    flush_bit_stream_outgoing_interface(i);

  /* Ok, now we're sure there is free space for the bit. Set it if it's 1
     (the buffer was cleared at initialization or at the last flush, so 
     there's no need to clear any bit when writing a 0): */
  if(bit){
#ifdef BUFFER_LENGTH_IN_WORDS
    const unsigned_integer_t word_index = i->next_bit_to_write / WORD_BIT;
    const unsigned_integer_t offset = i->next_bit_to_write % WORD_BIT;
    const unsigned_integer_t bitmask = 1u << offset;
    i->buffer[word_index] |= bitmask;
#else
    const unsigned_integer_t bitmask = 1u << i->next_bit_to_write;
    i->buffer |= bitmask;
#endif  
  } // if
  
  /* Advance the index referring the next bit to use: */
  i->next_bit_to_write++;
}

void read_bits(bit_stream_incoming_interface_t i,
               unsigned_integer_t *bits,
               const unsigned_integer_t bits_no){
  int j;
  for(j = 0u; j < bits_no; j++){
    const unsigned_integer_t word_index = j / WORD_BIT;
    const unsigned_integer_t offset = j % WORD_BIT;
    const unsigned_integer_t bitmask = 1u << offset;
    if(read_bit(i))
      bits[word_index] |= bitmask;
    else
      bits[word_index] &= ~bitmask;
  } // for
}

void write_bits(bit_stream_outgoing_interface_t i,
                const unsigned_integer_t *bits,
                const unsigned_integer_t bits_no){
  unsigned_integer_t j;
  for(j = 0u; j < bits_no; j++){
    const unsigned_integer_t word_index = j / WORD_BIT;
    const unsigned_integer_t offset = j % WORD_BIT;
    const unsigned_integer_t bitmask = 1u << offset;
    write_bit(i, bits[word_index] & bitmask);
  } // for
}

int test_write(){
  bit_stream_outgoing_interface_t out = malloc(1000);
  out->fd = STDERR_FILENO;
  out->next_bit_to_write = 0;

  DO_N_TIMES_BEGIN(1000){
    static unsigned_integer_t k = 0;
    write_bits(out, &k, 5);
    k++;
  } DO_TIMES_END;

  return 0;
}

int test_read(){
  bit_stream_incoming_interface_t in = malloc(1000);
  in->fd = STDIN_FILENO;
  in->next_bit_to_read = WORD_BIT;

  DO_N_TIMES_BEGIN(900){
    unsigned_integer_t h = 0u;
    read_bits(in, &h, 5);
    printf("%i\n", (int)h);
  } DO_TIMES_END;

  return 0;
}

int main(){
#define N 10000000u
  unsigned_integer_t array[N];
  memset(array, 0, N * sizeof(unsigned_integer_t));

  bit_stream_outgoing_interface_t out = malloc(100000);
  out->fd = STDERR_FILENO;
  out->next_bit_to_write = 0;

  int i;
#if 0
  TEMP_FAILURE_RETRY(write(out->fd, array, N * sizeof(unsigned_integer_t)));
#else
  write_bits(out, array, N * sizeof(unsigned_integer_t) * CHAR_BIT);
  flush_bit_stream_outgoing_interface(out);
#endif

  /*
  for(i = 0; i < N; i++){
#if 1
    TEMP_FAILURE_RETRY(write(out->fd, &array[i], sizeof(unsigned_integer_t)));
#else
    write_bits(out, &array[i], sizeof(unsigned_integer_t) * CHAR_BIT);
#endif
  } // for
  */
  return 0;

}
