/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EAM_LIBARIES_MEMORY_STACK_H_
#define EAM_LIBARIES_MEMORY_STACK_H_

#include <stdio.h>
#include "stack.h"
#include "../../common/types.h"
#include "../../common/malloc.h"
#include <assert.h>

/* A stack can be accessed either with a top pointer or a frame pointer; the
   caller may also use both, but in this case it's his/her responsibility to
   keep them coherent. In other words, no operation on the stack top pointer
   updates the frame pointer, and vice-versa. 
   
   The caller is free to store pointers in stacks, but it's not safe to use
   pointers to objects stored *within* the stack if the stack is resized,
   since resizing the stack may move all stores objects. Offsets should be
   stored instead.
   
   Bounds are checked only with assertions enabled, i.e. not in production
   runs. The caller should enstablish some reasonable policy for checking
   for overflow and resizing only when appropriate.

   This implementation of stacks makes use of a C struct (as such always
   kept in memory) holding data such as the pointer to the actual stack,
   allocated size, top and frame pointers. This is a covenient solution to
   adopt, but its performance is not optimal. See fast_stack.h for an
   alternative. */


/* All the following inline qualifiers are only expression of a wish with
   the current lack of link-time optimizations in GCC. Hopefully, the
   situation is going to change. */

/* Create a new stack with the given size. The size is in words: */
struct eam_memory_stack* eam_create_memory_stack(const unsigned_integer_t allocated_size)
  __attribute__(( malloc ));

/* Query/set the stack allocated size, always expressed in words. Resizing
   a stack automatically updates top and frame pointers. Be sure to see the
   initial comment about resizing in presence of pointers within the stack. */
inline unsigned_integer_t eam_get_memory_stack_allocated_size(const struct eam_memory_stack *stack)
  __attribute__(( pure ));
void eam_resize_memory_stack(const unsigned_integer_t new_size_in_words,
                      struct eam_memory_stack *stack);

/* Destroy the given stack: */
void eam_destroy_memory_stack(struct eam_memory_stack *stack);

/* Push/pop or query, using and updating the top pointer: */
inline void eam_push_word_onto_memory_stack(const word_t datum,
                              struct eam_memory_stack *stack);
inline word_t eam_pop_word_from_memory_stack(struct eam_memory_stack *stack);
inline word_t eam_top_word_of_memory_stack(const struct eam_memory_stack *stack)
  __attribute__(( const ));

/* Return the number of objects, from the bottom (included) to the top pointer
   (excluded): */
inline unsigned_integer_t eam_memory_stack_elements_no(
    const struct eam_memory_stack *stack)
  __attribute__(( const ));

/* Explicitly update the top pointer: not needed if only using push/pop. The
   offset (in words) may be zero, positive or negative: */
inline void eam_set_memory_stack_top_pointer(const word_t *new_top_pointer,
                               struct eam_memory_stack *stack);
inline void eam_move_memory_stack_top_pointer(const integer_t offset,
                                struct eam_memory_stack *stack);

/* Update the frame pointer: */
inline void eam_set_memory_stack_frame_pointer(const word_t *new_frame_pointer,
                                 struct eam_memory_stack *stack);
inline void eam_move_memory_stack_frame_pointer(const integer_t offset,
                                  struct eam_memory_stack *stack);

/* Query pointers: */
inline word_t* eam_get_memory_stack_top_pointer(const struct eam_memory_stack *stack)
  __attribute__(( pure ));
inline word_t* eam_get_memory_stack_frame_pointer(const struct eam_memory_stack *stack)
  __attribute__(( pure ));
inline word_t* eam_get_memory_stack_bottom_pointer(const struct eam_memory_stack *stack)
  __attribute__(( pure ));

/* The same, for multi-word buffers: */
inline void eam_push_buffer_onto_memory_stack(const word_t *data,
                                const unsigned_integer_t size,
                                struct eam_memory_stack *stack);
inline void eam_pop_buffer_from_memory_stack(const word_t *data,
                               const unsigned_integer_t size,
                               word_t *where_to_put_the_copy,
                               struct eam_memory_stack *stack);

/* Load/store, using the frame pointer. Offsets (in words) may be positive,
   zero or negative: */
inline word_t eam_load_word_from_memory_stack(const integer_t frame_pointer_offset,
                                struct eam_memory_stack *stack);
inline void eam_store_word_into_memory_stack(const integer_t frame_pointer_offset,
                               const word_t datum,
                               struct eam_memory_stack *stack);

/* The same, for multi-word buffers: */
inline void eam_load_buffer_from_memory_stack(const integer_t frame_pointer_offset,
                                const unsigned_integer_t size_in_words,
                                word_t *where_to_put_the_copy,
                                struct eam_memory_stack *stack);
inline void eam_store_buffer_into_memory_stack(const integer_t frame_pointer_offset,
                                 const word_t *data,
                                 const unsigned_integer_t size_in_words,
                                 struct eam_memory_stack *stack);

#endif // #ifndef EAM_LIBARIES_MEMORY_STACK_H_
