/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005 Luca Saiu
Updated in 2006 by Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#ifndef EAM_DRIVER_H_
#define EAM_DRIVER_H_

#include "../common/types.h"
#include "registers.h"

/* Blocks are implemented as C functions, returning continuations implemented
   as function pointers: */
#define EAM_BEGIN_EXPORTED_BLOCK(label) eam_continuation_t label(void){
#define EAM_BEGIN_INTERNAL_BLOCK(label) inline static eam_continuation_t label(void){
#ifdef ENABLE_ASSERTIONS_IN_GENERATED_CODE
#define EAM_END_BLOCK fprintf(stderr, "You should never reach this.\n"); assert(0);}
#else
#define EAM_END_BLOCK }
#endif // #ifdef ENABLE_ASSERTIONS_IN_GENERATED_CODE

/* Block declarations: */
#define EAM_DECLARE_EXPORTED_BLOCK(label) \
  eam_continuation_t label(void) __attribute__(( nothrow, flatten ))
#define EAM_DECLARE_INTERNAL_BLOCK(label) \
  inline static eam_continuation_t label(void) __attribute__(( nothrow, flatten ))

/* A jump to another block is implemented by returning a function pointer to
   the driver, which in turn calls the returned function: this costs slightly
   more than a pointer return plus a function pointer call. The "slightly more"
   part is due to the loop in the driver, which implies a jump (possibly not
   always, because of unrolling): */
#define EAM_SLOW_JUMP(label) return (eam_continuation_t)(label)

/* As a special case, a jump to a block within the current compilation unit (or
   to another compilation unit compiled in the same GCC command line) can be
   implemented as a tail call, because modern GCCs are able to optimize this 
   pattern. A local jump is compiled into a jump instruction or into an inline
   copy of the jumped-to block, hence it is *much* cheaper than a slow jump on
   modern architectures: */
#define EAM_FAST_JUMP(label) return ((eam_continuation_t(*)(void)) (label))()

/* Jump to a block using the "fast" method (a C function call) even if this could
   mean pushing on the C control stack. This is useful in cases when the GCC is
   conservative enough to refuse performing sibling call optimization, but we know
   that the C stack will not grow unbounded because only O(1) jumps will be made,
   and we want them to be fast. If GCC can optimize the sibling call in this case
   then this is exactly as fast as a FAST_JUMP. Of course this should be
   used with care. */
#define EAM_NON_TAIL_JUMP(label) EAM_FAST_JUMP(label)

/* Enter the eAM driver loop executing the given continuation. To exit the
   driver and return the user should  call eam_exit(). This weird control
   flow idiom is a non-local exit implemented via setjmp()/longjmp(). At most
   *one* eAM driver can be active at any given time. This restriction may be
   lifted in the future. The returned value is the exit code given to
   eam_exit(). */
integer_t eam_driver(const eam_continuation_t s);

/* Exit the eAM driver and return from that (implemented via setjmp()). This
   should only be called from within the driver, i.e. in eAM primitives. Calling
   this procedure has the effect to make eam_driver() return the value exit_code
   and give control back to the caller. */
void eam_exit(const integer_t exit_code)
  __attribute__((nothrow, flatten, noreturn));

/* Return (a safe approximation of) the C stack bottom; this is useful to
   initialize the GC, when needed. This is only usable within the driver loop,
   because what is returned is the address of a local object which is only
   alive when the driver loop is active. */
void* eam_get_stack_bottom(void)
  __attribute__(( pure ));

#endif // #ifndef EAM_DRIVER_H_
