/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#ifndef EAM_EAML_H_
#define EAM_EAML_H_

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "../common/types.h"
#include "../common/list.h"
#include "../common/mangle.h"

/* Here comes a simple C implementation of the eAML abstract syntax. */

struct eam_immediate{
  //char *type;
  char *value;
}; // struct

struct eam_primitive_call{
  char *operator;
  struct eam_list *operands;
}; // struct

struct eam_register{
  enum{
    eam_indexed_register_register_type,
    eam_named_register_register_type
  } type;
  union{
    integer_t register_no;
    char *register_name; // without the "%" prefix
  }; // anonymous union
}; // struct

struct eam_register_definition{
  enum{
    eam_indexed_register_register_definition_type,
    eam_memory_register_definition_type
  } type;
  union{
    integer_t register_no;
    // nothing is needed for the 'memory' case
  }; // anonymous union
}; // struct

struct eam_expression{
  enum{
    eam_immediate_expression_type,
    eam_register_expression_type,
    eam_label_expression_type,
    eam_primitive_call_expression_type,
    eam_global_expression_type,
  } type;
  union{
    struct eam_immediate *immediate;
    struct eam_register *register_; //The identifier "register" is reserved in C
    char *label;
    struct eam_primitive_call* primitive_call;
    char *global;
  }; // anonymous union
}; // struct

struct eam_command{
  enum eam_command_type{
    eam_label_command_type,
    eam_primitive_call_command_type,
    eam_assignment_command_type,
    eam_goto_command_type,
    non_tail_recursive_eam_goto_command_type,
    eam_if_command_type
  } type;
  union{
    struct{
      char *label;
      integer_t exported;
    } label_command;
    struct{
      struct eam_primitive_call* primitive_call;
    } primitive_call_command;
    struct{
      struct eam_expression* left_expression;
      struct eam_expression* right_expression;
    } assignment_command;
    struct{
      struct eam_expression* target_expression;
    } goto_command;
    struct{
      struct eam_expression* target_expression;
    } non_tail_recursive_goto_command;
    struct{
      struct eam_expression* guard_expression;
      struct eam_expression* target_expression;
    } if_command;
  }; // anonymous union
}; // struct

struct eam_primitive_definition{
  char *name;
  integer_t exported;
  integer_t arity;
  char *c_name;
  char *c_declaration;
  char *c_implementation;
}; // struct

struct eam_global_definition{
  char *name;
  integer_t exported;
  integer_t constant;
  // To do: allow initialization, otherwise "constant" is useless
}; // struct

struct eam_definition{
  enum eam_definition_type{
    eam_immediate_type_definition_type,
    eam_primitive_definition_type,
    eam_global_definition_type
  } type;
  union{
    /* To do: struct eam_immediate_type_definition is only forward-declared here as
       of now, and is neither defined nor used. */
    struct eam_immediate_type_definition* immediate_type_definition;
    struct eam_primitive_definition* primitive_definition;
    struct eam_global_definition* global_definition;
  }; // anonymous union
}; // struct

struct eam_initialization{
  char *c_declaration;
  char *c_implementation;
  char *c_procedure_body;
}; // struct

struct eam_finalization{
  char *c_declaration;
  char *c_implementation;
  char *c_procedure_body;
}; // struct

/* A structure holding together two strings, where the second one is the mangled
   version of the first: */
struct eam_unmangled_and_mangled{
  char *unmangled;
  char *mangled;
}; // struct

struct eam_register_naming{
  char *name; // without the "%" prefix
  struct eam_register_definition *definition;
}; // struct

/* A whole single compilation unit, the result of parsing an eAML file: */
struct eam_compilation_unit{
  char *name;
  char *mangled_name;
  struct eam_list *imports; // a list of struct eam_unmangled_and_mangled
  struct eam_list *non_eam_libraries; // a list of strings
  struct eam_list *register_namings; // a list of struct eam_register_naming
  struct eam_list *definitions; // a list of struct eam_definition
  struct eam_list *commands; // a list of struct eam_command
  struct eam_initialization *initialization;
  struct eam_finalization *finalization;
}; // struct

/* The parser driver public entry point. This returns a pointer to a
   new compilation_unit structure holding the abstract syntax tree, or NULL
   in case of error. If the given file name is "-" then stdin is used for
   input. Non-reentrant, but returning completely distinct structures with
   no sharing. Defined in eamlparser.y: */
struct eam_compilation_unit* eam_parse_compilation_unit(char *file_name);

/* Deallocate compilation_unit and all the data structures it points to; this
   assumes the strucure was created by the parser, which doesn't share any
   common substructure and malloc()s every pointed string, struct or union. */
void eam_destroy_compilation_unit(struct eam_compilation_unit *compilation_unit);

#endif // #ifndef EAM_EAML_H_
