/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005, 2006 Luca Saiu

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

%{
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "../common/list.h"
#include "../common/types.h"
#include "../common/malloc.h"
#include "eaml.h"
#include "../common/mangle.h"

extern int yylineno;
extern char* yytext;
extern int yyleng;
extern FILE *yyin, *yyout;

/* Ugly globals, set by the parser. Non-reentrant. */
static struct eam_compilation_unit *parsed_compilation_unit;
char* file_name_being_parsed  = "<unnamed-file>";

int yylex(void);
int yyerror(char*);

/* Return a malloc()ed version of ""; we want to be able to destroy *all* the
   strings generated by the parser with a free() call: */
static char* make_empty_string();
%}

%union{
  struct eam_expression *expression;
  struct eam_expression *left_expression;
  struct eam_register *register_;
  struct eam_command *command;
  struct eam_immediate_type *immediate_type;
  struct eam_global_definition *global_definition;
  struct eam_primitive_definition *primitive_definition;
  struct eam_compilation_unit *compilation_unit;
  struct eam_primitive_call *primitive_call;
  struct eam_list *list;
  char *label;
  char *global;
  struct eam_immediate *immediate;
  struct eam_definition *definition;
  struct eam_initialization *initialization;
  struct eam_finalization *finalization;
  integer_t boolean;
  char *string;
  integer_t number;
  struct eam_register_naming *register_naming;
  struct eam_register_definition *register_definition;
}

%nonassoc LOWEST_PRECEDENCE

%token COMPILATION_UNIT
%token IMPORT LIBRARY
%token INITIALIZATION FINALIZATION
%token CONSTANT PRIMITIVE CODE DEFINE_REGISTER AS MEMORY
%token ARITY C_NAME C_DECLARATION C_IMPLEMENTATION C_PROCEDURE_BODY END
%token OPENPAR CLOSEPAR COMMA
%token NAMED_REGISTER INDEXED_REGISTER LABEL
%token IDENTIFIER NUMBER EMBEDDED_CODE IMMEDIATE
%token EXPORTED
%token GLOBAL
%token SET TAKES
%token GOTO IF NON_TAIL_RECURSIVE_GOTO

%nonassoc HIGHEST_PRECEDENCE

%start compilation_unit

%type <immediate> immediate
%type <global> global
%type <label> label
%type <number> number
%type <string> string
%type <string> embedded_code
%type <expression> expression
%type <left_expression> left_expression
%type <command> command
%type <list> commands
%type <list> code
%type <list> imports
%type <list> non_eam_libraries
%type <register_> register_

%type <boolean> optional_exported
%type <boolean> optional_constant

 /* %type <immediate_type> immediate_type */
%type <global_definition> global_definition
%type <primitive_definition> primitive_definition

%type <compilation_unit> compilation_unit
%type <string> optional_c_declaration
%type <string> optional_c_implementation
%type <string> optional_c_procedure_body
%type <initialization> optional_initialization
%type <finalization> optional_finalization
%type <primitive_call> primitive_call
%type <definition> definition
%type <list> definitions
%type <list> zero_or_more_arguments
%type <list> one_or_more_arguments
%type <list> more_arguments

%type <string> named_register 
%type <list> register_namings
%type <register_naming> register_naming
%type <register_definition> register_definition

%%

compilation_unit :
  COMPILATION_UNIT string
  imports
  non_eam_libraries
  register_namings
  optional_initialization
  optional_finalization
  definitions
  code
  END COMPILATION_UNIT
  { $$ = (struct eam_compilation_unit*)eam_xmalloc(sizeof(struct eam_compilation_unit));
    $$->name = $2;
    $$->mangled_name =
      eam_mangle_with_limit($2, EAML_COMPILATION_UNIT_MANGLED_NAME_LENGTH);
    $$->imports = $3;
    $$->non_eam_libraries = $4;
    $$->register_namings = $5;
    $$->initialization = $6;
    $$->finalization = $7;
    $$->definitions = $8;
    $$->commands = $9;
    /* Update the global pointer a copy of which returned by the parser
       driver: */
    parsed_compilation_unit = $$; }
;

code :
  /* nothing */
  { $$ = eam_nil; }
| CODE commands END CODE
  { $$ = $2; }
;

optional_initialization :
  /* nothing */
  { $$ = NULL; }
| INITIALIZATION
  optional_c_declaration
  optional_c_implementation
  optional_c_procedure_body
  END INITIALIZATION
  { $$ = (struct eam_initialization*)eam_xmalloc(sizeof(struct eam_initialization));
    $$->c_declaration = $2;
    $$->c_implementation = $3; 
    $$->c_procedure_body = $4; }
;

optional_finalization :
  /* nothing */
  { $$ = NULL; }
| FINALIZATION
  optional_c_declaration
  optional_c_implementation
  optional_c_procedure_body
  END FINALIZATION
  { $$ = (struct eam_finalization*)eam_xmalloc(sizeof(struct eam_finalization));
    $$->c_declaration = $2;
    $$->c_implementation = $3; 
    $$->c_procedure_body = $4; }
;

non_eam_libraries:
  /* nothing */
  { $$ = eam_nil; }
| LIBRARY string non_eam_libraries
  { $$ = eam_cons($2, $3); }

imports :
  /* nothing*/
  { $$ = eam_nil; }
| IMPORT string imports
  { struct eam_unmangled_and_mangled *pair =
      (struct eam_unmangled_and_mangled*)eam_xmalloc(sizeof(struct eam_unmangled_and_mangled));
    pair->unmangled = $2;
    pair->mangled =
      eam_mangle_with_limit($2, EAML_COMPILATION_UNIT_MANGLED_NAME_LENGTH);
    $$ = eam_cons(pair, $3); }
;

optional_c_declaration :
  /* nothing */
  { $$ = make_empty_string(); }
| C_DECLARATION embedded_code
  { $$ = $2; }
;

optional_c_implementation :
  /* nothing */
  { $$ = make_empty_string(); }
| C_IMPLEMENTATION embedded_code
  { $$ = $2; }
;

optional_c_procedure_body :
  /* nothing */
  { $$ = make_empty_string(); }
| C_PROCEDURE_BODY embedded_code
  { $$ = $2; }
;

definition :
/*  immediate_type |*/
  global_definition
  { $$ = (struct eam_definition*)eam_xmalloc(sizeof(struct eam_definition));
    $$->type = eam_global_definition_type;
    $$->global_definition = $1; }
| primitive_definition
  { $$ = (struct eam_definition*)eam_xmalloc(sizeof(struct eam_definition));
    $$->type = eam_primitive_definition_type;
    $$->primitive_definition = $1; }
;

definitions :
  /* nothing */
  { $$ = eam_nil; }
| definition definitions
  { $$ = eam_cons($1, $2); }
;

global_definition :
  optional_exported
  optional_constant
  GLOBAL
  string
  { $$ = (struct eam_global_definition*)eam_xmalloc(sizeof(struct eam_global_definition));
    $$->exported = $1;
    $$->constant = $2;
    $$->name = $4; }
;

optional_exported :
  /* nothing*/
  { $$ = 0; }
| EXPORTED
  { $$ = 1; }
;

optional_constant :
  /* nothing*/
  { $$ = 0; }
| CONSTANT
  { $$ = 1; }
;

primitive_definition :
  optional_exported PRIMITIVE string
  ARITY number
  C_NAME string
  C_DECLARATION embedded_code
  C_IMPLEMENTATION embedded_code
  END PRIMITIVE
  { $$ = (struct eam_primitive_definition*)
      eam_xmalloc(sizeof(struct eam_primitive_definition));
    $$->exported = $1;
    $$->name = $3;
    $$->arity = $5;
    $$->c_name = $7;
    $$->c_declaration = $9;
    $$->c_implementation = $11; }
;

register_namings :
  /* nothing */
  { $$ = eam_nil; }
| register_naming register_namings
  { $$ = eam_cons($1, $2); }
;

register_naming :
  DEFINE_REGISTER named_register AS register_definition
  { $$ = (struct eam_register_naming*)
      eam_xmalloc(sizeof(struct eam_register_naming));
    $$->name = $2;
    //fprintf(stderr, "[%s][%s]\n", $$->name, $2);
    $$->definition = $4; }
;

named_register : 
  NAMED_REGISTER
  { $$ = eam_xmalloc(yyleng + 1); 
    strcpy($$, yytext); }

register_definition :
  MEMORY
  { $$ = (struct eam_register_definition*)
      eam_xmalloc(sizeof(struct eam_register_definition));
    $$->type = eam_memory_register_definition_type; }
| INDEXED_REGISTER
  { $$ = (struct eam_register_definition*)
      eam_xmalloc(sizeof(struct eam_register_definition));
    int n; sscanf(yytext, "%%r%i", &n);
    $$->type = eam_indexed_register_register_definition_type;
    $$->register_no = (integer_t)n; }
;

command :
  optional_exported label
  { $$ = (struct eam_command*)eam_xmalloc(sizeof(struct eam_command));
    $$->type = eam_label_command_type;
    $$->label_command.label = $2;
    $$->label_command.exported = $1; }
| primitive_call
  { $$ = (struct eam_command*)eam_xmalloc(sizeof(struct eam_command));
    $$->type = eam_primitive_call_command_type;
    $$->primitive_call_command.primitive_call = $1; }
| left_expression TAKES expression
  { $$ = (struct eam_command*)eam_xmalloc(sizeof(struct eam_command));
    $$->type = eam_assignment_command_type;
    $$->assignment_command.left_expression = $1;
    $$->assignment_command.right_expression = $3; }
| GOTO expression
  { $$ = (struct eam_command*)eam_xmalloc(sizeof(struct eam_command));
    $$->type = eam_goto_command_type;
    $$->goto_command.target_expression = $2; }
| NON_TAIL_RECURSIVE_GOTO expression
  { $$ = (struct eam_command*)eam_xmalloc(sizeof(struct eam_command));
    $$->type = non_tail_recursive_eam_goto_command_type;
    $$->non_tail_recursive_goto_command.target_expression = $2; }
| IF expression GOTO expression
  { $$ = (struct eam_command*)eam_xmalloc(sizeof(struct eam_command));
    $$->type = eam_if_command_type;
    $$->if_command.guard_expression = $2; 
    $$->if_command.target_expression = $4; }
;

commands :
  /* nothing */ 
  { $$ = eam_nil; }
| command commands
  { $$ = eam_cons($1, $2); }
;

number :
  NUMBER
  { int n; sscanf(yytext, "%i", &n); $$ = (integer_t)n; }
;

register_ :
  INDEXED_REGISTER
  { int n; sscanf(yytext, "%%r%i", &n);
    $$ = (struct eam_register*)eam_xmalloc(sizeof(struct eam_register));
    $$->type = eam_indexed_register_register_type;
    $$->register_no = (integer_t)n; }
|
  NAMED_REGISTER
  { $$ = (struct eam_register*)eam_xmalloc(sizeof(struct eam_register));
    $$->type = eam_named_register_register_type;
    $$->register_name = (char*)eam_xmalloc(yyleng + 1);
    strcpy($$->register_name, yytext); }
;

label :
  LABEL
  { char *buffer = (char*)eam_xmalloc(yyleng + 1);
    strcpy(buffer, yytext);
    $$ = buffer; }
;

immediate :
  IMMEDIATE OPENPAR string CLOSEPAR
  { $$ = (struct eam_immediate*)eam_xmalloc(sizeof(struct eam_immediate));
    $$->value = $3; }
;

global :
  GLOBAL OPENPAR string CLOSEPAR
  { $$ = $3; }
;

string :
  IDENTIFIER
  { char *buffer = (char*)eam_xmalloc(yyleng + 1);
    strcpy(buffer, yytext);
    $$ = buffer; }
| NUMBER
  { char *buffer = (char*)eam_xmalloc(yyleng + 1);
    strcpy(buffer, yytext);
    $$ = buffer; }
;

embedded_code:
  EMBEDDED_CODE
  { char *buffer = (char*)eam_xmalloc(yyleng + 1);
    strcpy(buffer, yytext);
    $$ = buffer; }
;

left_expression :
  register_
  { $$ = (struct eam_expression*)eam_xmalloc(sizeof(struct eam_expression)); 
    $$->type = eam_register_expression_type;
    $$->register_ = $1; }
| global
  { $$ = (struct eam_expression*)eam_xmalloc(sizeof(struct eam_expression)); 
    $$->type = eam_global_expression_type;
    $$->global = $1; }
| primitive_call
  { $$ = (struct eam_expression*)eam_xmalloc(sizeof(struct eam_expression));
    $$->type = eam_primitive_call_expression_type;
    $$->primitive_call = $1; }
| OPENPAR left_expression CLOSEPAR
  { $$ = $2; }
;

expression :
  immediate
  { $$ = (struct eam_expression*)eam_xmalloc(sizeof(struct eam_expression)); 
    $$->type = eam_immediate_expression_type;
    $$->immediate = $1; }
| register_
  { $$ = (struct eam_expression*)eam_xmalloc(sizeof(struct eam_expression)); 
    $$->type = eam_register_expression_type;
    $$->register_ = $1; }
| label
  { $$ = (struct eam_expression*)eam_xmalloc(sizeof(struct eam_expression)); 
    $$->type = eam_label_expression_type;
    $$->label = $1; }
| global
  { $$ = (struct eam_expression*)eam_xmalloc(sizeof(struct eam_expression));
    $$->type = eam_global_expression_type;
    $$->global = $1; }
| primitive_call
  { $$ = (struct eam_expression*)eam_xmalloc(sizeof(struct eam_expression));
    $$->type = eam_primitive_call_expression_type;
    $$->primitive_call = $1; }
| OPENPAR expression CLOSEPAR
  { $$ = $2; }
;

primitive_call :
  string OPENPAR zero_or_more_arguments
  { $$ = (struct eam_primitive_call*)eam_xmalloc(sizeof(struct eam_primitive_call));
    $$->operator = $1;
    $$->operands = $3; }
;

zero_or_more_arguments :
  CLOSEPAR
  { $$ = eam_nil; }
| COMMA CLOSEPAR /* we also accept a final comma before ")" */
  { $$ = eam_nil; }
| one_or_more_arguments
  { $$ = $1; }
;

one_or_more_arguments :
  expression more_arguments
  { $$ = eam_cons($1, $2); }
;


more_arguments :
  CLOSEPAR
  { $$ = eam_nil; }
| COMMA CLOSEPAR /* we accept a final comma before ")" */
  { $$ = eam_nil; }
| COMMA expression more_arguments
  { $$ = eam_cons($2, $3); }
;

%%

/* Make a fresh malloc()ed empty string: */
static char* make_empty_string(){
  char *r = (char*)eam_xmalloc(1);
  r[0] = '\0';
  return r;
}

/* The parser driver, non-reentrant but at least hiding side effects: */
struct eam_compilation_unit* eam_parse_compilation_unit(char *file_name){
  /* Is the given file name "-", standing for the standard input? */
  const int use_stdin = ! strcmp(file_name, "-");

  /* Save a pointer to the given filename into a global. This is used by
     yyerror(), which can be called by yyparse(). The lifetime of file_name
     is always enough for yyerror(): */
  file_name_being_parsed = file_name;  

  /* If we're not using stding then try to open the file, and fail 
     immediately if we can't; otherwise just use stdin: */
  if(use_stdin)
    yyin = stdin;
  else{
    yyin = fopen(file_name, "r");
    if(yyin == 0)
      return NULL;
  } // else
  
  /* Parse the file we opened: */
#warning To do: this leaks on parse error. Doing anything else is difficult,
#warning and probably not worth the effort. Should we exit with a fatal error
#warning on parse errors? The right thing would be to use the GC.
  yyparse();
  
  /* Ok, parsing succeeded. Close the file (unless it was stdin) and return
     the structure we built: */
  if(! use_stdin)
    fclose(yyin);
  return parsed_compilation_unit;
}
