/* This file is part of GNU epsilon, a functional language implementation

Copyright (C) 2005 Luca Saiu
Updated by Luca Saiu in 2006

GNU epsilon is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published
by the Free Software Foundation; either version 2, or (at your
option) any later version.

GNU epsilon is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU epsilon; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include "../common/types.h"
#include "../common/environment.h"
#include "eaml.h"
#include "symbol_table.h"
#include "eamlparser.h"
#include "generate_c.h"
#include "import_graph.h"
#include "../common/filesystem.h"
#include "../common/command_line.h"
#include "../common/malloc.h"

static void manage_command_line(int *argc, char **argv){
  /* Setup the "generic" part of command line support: */
  eam_set_program_name("eam-generate-c (" PACKAGE_NAME ")");
  eam_set_general_help_message("Compiles the given files.");
  eam_set_synopsis_string("eam-generate-c [OPTIONS] FILE1 ... FILEn");
  eam_set_version_string(VERSION_STRING);
  eam_set_copyright_string(COPYRIGHT_STRING);
  eam_set_license_message(LICENSE_STRING);
  eam_set_bug_reporting_message(BUG_REPORTING_MESSAGE);
  
  /* Add specific options: */
  eam_add_option_with_nonoptional_parameter("path",
					'p',
					".",//eam_getenv_or_empty_string("EAM_COMPILATION_UNIT_PATH"),
					"compilation unit search path");
  eam_add_option_with_nonoptional_parameter("symbol-path",
					's',
					".",//eam_getenv_or_empty_string("EAM_SYMBOL_TABLE_PATH"),
					"symbol table search path");
  
  
  /* Manage the actual command line: */
  eam_parse_command_line(argc, argv);

  //eam_show_options_status();
}



int main(int argc, char **argv){
  /* Support command line options: */
  manage_command_line (&argc, argv);
  
  //struct eam_list *search_path;
  //char *search_path_as_string;
  
  /* Set the search path: */
  char *search_path_as_string = eam_long_option_to_value("path");
  struct eam_list *search_path;
  search_path = eam_parse_search_path(search_path_as_string);
  
/*   /\* Set the search path: *\/ */
/*   search_path_as_string = getenv("EAMLPATH"); */
/*   if(search_path_as_string == NULL) */
/*     search_path_as_string = ""; */
/*   search_path = eam_cons(".", eam_parse_search_path(search_path_as_string)); */
  
  /* Parse all the given compilation units and add symbols to the symbol table: */
  struct eam_list *compilation_units = eam_nil;
  struct eam_symbol_table *symbol_table = eam_create_symbol_table();
  int i;
  for(i = 1; i < argc; i++){
    /* Lookup the file name: */
    char *file_name = eam_locate_in_search_path(argv[i], search_path);
    if(file_name == NULL)
      eam_fatal("could not find %s in the search path %s", argv[i], search_path_as_string);
    
    //fprintf(stderr, "Parsing %s... ", file_name);
    struct eam_compilation_unit *compilation_unit = eam_parse_compilation_unit(file_name);
    compilation_units = eam_cons(compilation_unit, compilation_units);
    //fprintf(stderr, "ok, the compilation unit is at %p.\n", compilation_unit);

    //fprintf(stderr, "Adding %s's symbols to the symbol table... ", file_name);
    eam_fill_symbol_table_from_compilation_unit(symbol_table,
					    compilation_unit,
					    0);
    free(file_name);
    //fprintf(stderr, "ok.\n");
  } // for
  
  /* To do: it is not a good idea to have two distinct symbol tables. Symbol
     tables are useful to catch error such as duplicate definitions...
     Of course we should use a repository for this, as soon as repositories
     are implementd. */
  struct eam_symbol_table *library_symbol_table = eam_create_symbol_table();

  /* Build the import graph and check there are no cycles: */
  //fprintf(stderr, "Building the import graph... ");
  struct eam_import_graph *import_graph =
    eam_create_import_graph(compilation_units, library_symbol_table);
  //fprintf(stderr, "ok, it's at %p.\n", import_graph);
  
  /* Make a temporary directory: */
  const char *temporary_directory = "output";//eam_create_temporary_directory();
  if(temporary_directory == NULL)
    eam_fatal("could not create a temporary directory");
  /*
  fprintf(stderr, "Creating files in the temporary directory %s...\n",
	  temporary_directory);
  */
  //chdir(temporary_directory);
  
  /* Generate C files in the temporary directory: */
  const struct eam_list *p = compilation_units;
  for(p = compilation_units; ! eam_null(p); p = eam_cdr(p)){
    struct eam_compilation_unit *compilation_unit = (struct eam_compilation_unit *)eam_car(p);
    //fprintf(stderr, "Generating C code from %s... ", compilation_unit->name);
    eam_generate_c_file(compilation_unit,
		    symbol_table,
		    temporary_directory);
    //fprintf(stderr, "ok.\n");
  }
  
  /* Generate a single C file #including the other ones: */
  eam_generate_single_including_c_file(compilation_units, temporary_directory);

  /* Generate a set of C files #including the other ones: */
  integer_t sets_no = eam_generate_including_c_files(compilation_units,
                                                     temporary_directory);
  //  fprintf(stderr, "%i sets were generated.\n", sets_no);
  
#ifdef ENABLE_ASSERTIONS
  /* Just a robustness test: it serves no other purpose, since we're exiting: */
  eam_destroy_import_graph(import_graph);
  eam_destroy_symbol_table(symbol_table);
  eam_destroy_symbol_table(library_symbol_table);
#endif // #ifdef ENABLE_ASSERTIONS
  
  /* We're done. */
  return EXIT_SUCCESS;
}
